from distutils.version import LooseVersion
import functools

import torch


def no_grad():
    version = torch.__version__ if str(torch.__version__) is torch.__version__ else "0.4.1"
    if LooseVersion(version) < LooseVersion("0.4.1"):  # No grad isn't a decorator
        def decorator(func):
            @functools.wraps(func)
            def wrap_no_grad(*args, **kwargs):
                with torch.no_grad():
                    return func(*args, **kwargs)
            return wrap_no_grad
        return decorator
    else:
        return torch.no_grad()


class Metric(object):
    """Base metric class. Process will be called on each batch, process-final at the end of each epoch.
    The metric contract allows for metrics to take any args but not kwargs. The initial metric call will be given state,
    however, subsequent metrics can pass any values desired.

    .. note::

        All metrics must extend this class.

    Args:
        name (str): The name of the metric
    """

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    @no_grad()
    def process(self, *args):
        """Process the state and update the metric for one iteration.

        Args:
            args: Arguments given to the metric. If this is a root level metric, will be given state

        Returns:
            None, or the value of the metric for this batch
        """
        pass

    @no_grad()
    def process_final(self, *args):
        """Process the terminal state and output the final value of the metric.

        Args:
            args: Arguments given to the metric. If this is a root level metric, will be given state

        Returns:
            None or the value of the metric for this epoch
        """
        pass

    def eval(self, data_key=None):
        """Put the metric in eval mode during model validation.
        """
        pass

    def train(self):
        """Put the metric in train mode during model training.
        """
        pass

    def reset(self, state):
        """Reset the metric, called before the start of an epoch.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass


class Callback(object):
    """Base callback class.

    .. note::

        All callbacks should override this class.

    """

    def state_dict(self):
        """Get a dict containing the callback state.

        Returns:
            dict: A dict containing parameters and persistent buffers.
        """
        return {}

    def __str__(self):
        return str(self.__class__).replace('<class ', '').replace('>', '').replace("'", "")

    def load_state_dict(self, state_dict):
        """Resume this callback from the given state. Expects that this callback was constructed in the same way.

        Args:
            state_dict (dict): The state dict to reload

        Returns:
            :class:`.Callback`: self
        """
        return self

    def on_start(self, state):
        """Perform some action with the given state as context at the start of a model fit.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_start_epoch(self, state):
        """Perform some action with the given state as context at the start of each epoch.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_start_training(self, state):
        """Perform some action with the given state as context at the start of the training loop.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_sample(self, state):
        """Perform some action with the given state as context after data has been sampled from the generator.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_forward(self, state):
        """Perform some action with the given state as context after the forward pass (model output) has been completed.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_criterion(self, state):
        """Perform some action with the given state as context after the criterion has been evaluated.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_backward(self, state):
        """Perform some action with the given state as context after backward has been called on the loss.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_step_training(self, state):
        """Perform some action with the given state as context after step has been called on the optimiser.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_end_training(self, state):
        """Perform some action with the given state as context after the training loop has completed.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_end_epoch(self, state):
        """Perform some action with the given state as context at the end of each epoch.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_checkpoint(self, state):
        """Perform some action with the state after all other callbacks have completed at the end of an epoch and the
        history has been updated. Should only be used for taking checkpoints or snapshots and will only be called by the
        run method of Trial.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_end(self, state):
        """Perform some action with the given state as context at the end of the model fitting.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_start_validation(self, state):
        """Perform some action with the given state as context at the start of the validation loop.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_sample_validation(self, state):
        """Perform some action with the given state as context after data has been sampled from the validation generator.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_forward_validation(self, state):
        """Perform some action with the given state as context after the forward pass (model output) has been completed
        with the validation data.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_criterion_validation(self, state):
        """Perform some action with the given state as context after the criterion evaluation has been completed
        with the validation data.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_end_validation(self, state):
        """Perform some action with the given state as context at the end of the validation loop.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_step_validation(self, state):
        """Perform some action with the given state as context at the end of each validation step.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass



from unittest import TestCase
from mock import patch, Mock

import torchbearer
from torchbearer.callbacks import TorchScheduler, LambdaLR, StepLR, MultiStepLR, ExponentialLR, CosineAnnealingLR,\
    ReduceLROnPlateau


class TestTorchScheduler(TestCase):
    def test_torch_scheduler_on_batch_with_monitor(self):
        state = {torchbearer.EPOCH: 1, torchbearer.METRICS: {'test': 101}, torchbearer.OPTIMIZER: 'optimizer'}
        mock_scheduler = Mock()
        mock_scheduler.return_value = mock_scheduler

        torch_scheduler = TorchScheduler(lambda opt: mock_scheduler(opt), monitor='test', step_on_batch=True)

        torch_scheduler.on_start(state)
        mock_scheduler.assert_called_once_with('optimizer')
        mock_scheduler.reset_mock()

        torch_scheduler.on_start_training(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()

        torch_scheduler.on_sample(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()

        torch_scheduler.on_step_training(state)
        mock_scheduler.step.assert_called_once_with(101)
        mock_scheduler.reset_mock()

        torch_scheduler.on_end_epoch(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()

    def test_torch_scheduler_on_epoch_with_monitor(self):
        state = {torchbearer.EPOCH: 1, torchbearer.METRICS: {'test': 101}, torchbearer.OPTIMIZER: 'optimizer'}
        mock_scheduler = Mock()
        mock_scheduler.return_value = mock_scheduler

        torch_scheduler = TorchScheduler(lambda opt: mock_scheduler(opt), monitor='test', step_on_batch=False)

        torch_scheduler.on_start(state)
        mock_scheduler.assert_called_once_with('optimizer')
        mock_scheduler.reset_mock()

        torch_scheduler.on_start_training(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()

        torch_scheduler.on_sample(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()

        torch_scheduler.on_step_training(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()

        torch_scheduler.on_end_epoch(state)
        mock_scheduler.step.assert_called_once_with(101, epoch=1)
        mock_scheduler.reset_mock()

    def test_torch_scheduler_on_batch_no_monitor(self):
        state = {torchbearer.EPOCH: 1, torchbearer.OPTIMIZER: 'optimizer'}
        mock_scheduler = Mock()
        mock_scheduler.return_value = mock_scheduler

        torch_scheduler = TorchScheduler(lambda opt: mock_scheduler(opt), monitor=None, step_on_batch=True)

        torch_scheduler.on_start(state)
        mock_scheduler.assert_called_once_with('optimizer')
        mock_scheduler.reset_mock()

        torch_scheduler.on_start_training(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()

        torch_scheduler.on_sample(state)
        mock_scheduler.step.assert_called_once_with()
        mock_scheduler.reset_mock()

        torch_scheduler.on_step_training(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()

        torch_scheduler.on_end_epoch(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()

    def test_torch_scheduler_on_epoch_no_monitor(self):
        state = {torchbearer.EPOCH: 1, torchbearer.OPTIMIZER: 'optimizer'}
        mock_scheduler = Mock()
        mock_scheduler.return_value = mock_scheduler

        torch_scheduler = TorchScheduler(lambda opt: mock_scheduler(opt), monitor=None, step_on_batch=False)

        torch_scheduler.on_start(state)
        mock_scheduler.assert_called_once_with('optimizer')
        mock_scheduler.reset_mock()

        torch_scheduler.on_start_training(state)
        mock_scheduler.step.assert_called_once_with(epoch=1)
        mock_scheduler.reset_mock()

        torch_scheduler.on_sample(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()

        torch_scheduler.on_step_training(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()

        torch_scheduler.on_end_epoch(state)
        mock_scheduler.assert_not_called()
        mock_scheduler.reset_mock()


class TestLambdaLR(TestCase):
    @patch('torch.optim.lr_scheduler.LambdaLR')
    def test_lambda_lr(self, lr_mock):
        state = {torchbearer.OPTIMIZER: 'optimizer'}

        scheduler = LambdaLR(0.1, last_epoch=-4, step_on_batch='batch')
        scheduler.on_start(state)

        lr_mock.assert_called_once_with('optimizer', 0.1, last_epoch=-4)
        self.assertTrue(scheduler._step_on_batch == 'batch')


class TestStepLR(TestCase):
    @patch('torch.optim.lr_scheduler.StepLR')
    def test_lambda_lr(self, lr_mock):
        state = {torchbearer.OPTIMIZER: 'optimizer'}

        scheduler = StepLR(10, gamma=0.4, last_epoch=-4, step_on_batch='batch')
        scheduler.on_start(state)

        lr_mock.assert_called_once_with('optimizer', 10, gamma=0.4, last_epoch=-4)
        self.assertTrue(scheduler._step_on_batch == 'batch')


class TestMultiStepLR(TestCase):
    @patch('torch.optim.lr_scheduler.MultiStepLR')
    def test_lambda_lr(self, lr_mock):
        state = {torchbearer.OPTIMIZER: 'optimizer'}

        scheduler = MultiStepLR(10, gamma=0.4, last_epoch=-4, step_on_batch='batch')
        scheduler.on_start(state)

        lr_mock.assert_called_once_with('optimizer', 10, gamma=0.4, last_epoch=-4)
        self.assertTrue(scheduler._step_on_batch == 'batch')


class TestExponentialLR(TestCase):
    @patch('torch.optim.lr_scheduler.ExponentialLR')
    def test_lambda_lr(self, lr_mock):
        state = {torchbearer.OPTIMIZER: 'optimizer'}

        scheduler = ExponentialLR(0.4, last_epoch=-4, step_on_batch='batch')
        scheduler.on_start(state)

        lr_mock.assert_called_once_with('optimizer', 0.4, last_epoch=-4)
        self.assertTrue(scheduler._step_on_batch == 'batch')


class TestCosineAnnealingLR(TestCase):
    @patch('torch.optim.lr_scheduler.CosineAnnealingLR')
    def test_lambda_lr(self, lr_mock):
        state = {torchbearer.OPTIMIZER: 'optimizer'}

        scheduler = CosineAnnealingLR(4, eta_min=10, last_epoch=-4, step_on_batch='batch')
        scheduler.on_start(state)

        lr_mock.assert_called_once_with('optimizer', 4, eta_min=10, last_epoch=-4)
        self.assertTrue(scheduler._step_on_batch == 'batch')


class TestReduceLROnPlateau(TestCase):
    @patch('torch.optim.lr_scheduler.ReduceLROnPlateau')
    def test_lambda_lr(self, lr_mock):
        state = {torchbearer.OPTIMIZER: 'optimizer'}

        scheduler = ReduceLROnPlateau(monitor='test', mode='max', factor=0.2, patience=100, verbose=True, threshold=10,
                                      threshold_mode='thresh', cooldown=5, min_lr=0.1, eps=1e-4, step_on_batch='batch')
        scheduler.on_start(state)

        lr_mock.assert_called_once_with('optimizer', mode='max', factor=0.2, patience=100, verbose=True, threshold=10,
                                        threshold_mode='thresh', cooldown=5, min_lr=0.1, eps=1e-4)
        self.assertTrue(scheduler._step_on_batch == 'batch')
        self.assertTrue(scheduler._monitor == 'test')



from __future__ import print_function
import torchbearer

from torchbearer.callbacks import Callback

import math


class TerminateOnNaN(Callback):
    """Callback which montiors the given metric and halts training if its value is nan or inf.

    Args:
        monitor (str): The name of the metric to monitor

    State Requirements:
        - :attr:`torchbearer.state.METRICS`: Metrics should be a dict containing at least the key `monitor`
    """
    def __init__(self, monitor='running_loss'):
        super(TerminateOnNaN, self).__init__()
        self._monitor = monitor

    def _check(self, state):
        if self._monitor in state[torchbearer.METRICS]:
            value = state[torchbearer.METRICS][self._monitor]
            if value is not None:
                if math.isnan(value) or math.isinf(value):
                    print('Invalid ' + self._monitor + ', terminating')
                    state[torchbearer.STOP_TRAINING] = True

    def on_step_training(self, state):
        self._check(state)

    def on_end_epoch(self, state):
        self._check(state)

    def on_step_validation(self, state):
        self._check(state)



"""
    .. autoclass:: RocAucScore(one_hot_labels=True, one_hot_offset=0, one_hot_classes=10)
"""
from .decorators import default_for_key, to_dict
from .wrappers import EpochLambda

old_super = super


def super(_, obj):
    return old_super(obj.__class__, obj)


@default_for_key('roc_auc')
@default_for_key('roc_auc_score')
@to_dict
class RocAucScore(EpochLambda):
    """Area Under ROC curve metric. Default for keys: 'roc_auc', 'roc_auc_score'.

    .. note::

        Requires :mod:`sklearn.metrics`.

    Args:
        one_hot_labels (bool): If True, convert the labels to a one hot encoding. Required if they are not already.
        one_hot_offset (int): Subtracted from class labels, use if not already zero based.
        one_hot_classes (int): Number of classes for the one hot encoding.
    """

    def __init__(self, one_hot_labels=True, one_hot_offset=0, one_hot_classes=10):
        import sklearn.metrics
        import numpy as np

        def to_categorical(y):
            return np.eye(one_hot_classes, dtype='uint8')[y - one_hot_offset]

        if one_hot_labels:
            process = to_categorical
        else:
            process = lambda y: y

        super(RocAucScore, self).__init__('roc_auc_score', lambda y_pred, y_true: sklearn.metrics.roc_auc_score(process(y_true.cpu().numpy()), y_pred.cpu().detach().numpy()))



import sys
if sys.version_info[0] < 3:
    import inspect

    def count_args(fcn):
        return len(inspect.getargspec(fcn).args)
else:
    from inspect import signature

    def count_args(fcn):
        return len(signature(fcn).parameters)

import types

import torchbearer
from torchbearer.callbacks import Callback


class LambdaCallback(Callback):
    def __init__(self, func):
        self.func = func

    def on_lambda(self, state):
        return self.func(state)


def bind_to(target):
    def decorator(func):
        if isinstance(func, LambdaCallback):
            callback = func
        else:
            callback = LambdaCallback(func)
        setattr(callback, target.__name__, types.MethodType(lambda self, state: self.on_lambda(state), callback))
        return callback
    return decorator


def on_start(func):
    """ The :func:`on_start` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_start`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_start` calling func
    """
    return bind_to(Callback.on_start)(func)


def on_start_epoch(func):
    """ The :func:`on_start_epoch` decorator is used to initialise a :class:`.Callback` with
    :meth:`~.Callback.on_start_epoch` calling the decorated function

        Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_start_epoch` calling func
    """
    return bind_to(Callback.on_start_epoch)(func)


def on_start_training(func):
    """ The :func:`on_start_training` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_start_training`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_start_training` calling func
    """
    return bind_to(Callback.on_start_training)(func)


def on_sample(func):
    """ The :func:`on_sample` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_sample`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_sample` calling func
    """
    return bind_to(Callback.on_sample)(func)


def on_forward(func):
    """ The :func:`on_forward` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_forward`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_forward` calling func
    """
    return bind_to(Callback.on_forward)(func)


def on_criterion(func):
    """ The :func:`on_criterion` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_criterion`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_criterion` calling func
    """
    return bind_to(Callback.on_criterion)(func)


def on_backward(func):
    """ The :func:`on_backward` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_backward`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_backward` calling func
    """
    return bind_to(Callback.on_backward)(func)


def on_step_training(func):
    """ The :func:`on_step_training` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_step_training`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_step_training` calling func
    """
    return bind_to(Callback.on_step_training)(func)


def on_end_training(func):
    """ The :func:`on_end_training` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end_training`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_end_training` calling func
    """
    return bind_to(Callback.on_end_training)(func)


def on_end_epoch(func):
    """ The :func:`on_end_epoch` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end_epoch`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_end_epoch` calling func
    """
    return bind_to(Callback.on_end_epoch)(func)


def on_end(func):
    """ The :func:`on_end` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_end` calling func
    """
    return bind_to(Callback.on_end)(func)


def on_start_validation(func):
    """ The :func:`on_start_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_start_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_start_validation` calling func
    """
    return bind_to(Callback.on_start_validation)(func)


def on_sample_validation(func):
    """ The :func:`on_sample_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_sample_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_sample_validation` calling func
    """
    return bind_to(Callback.on_sample_validation)(func)


def on_forward_validation(func):
    """ The :func:`on_forward_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_forward_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_forward_validation` calling func
    """
    return bind_to(Callback.on_forward_validation)(func)


def on_criterion_validation(func):
    """ The :func:`on_criterion_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_criterion_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_criterion_validation` calling func
    """
    return bind_to(Callback.on_criterion_validation)(func)


def on_end_validation(func):
    """ The :func:`on_end_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_end_validation` calling func
    """
    return bind_to(Callback.on_end_validation)(func)


def on_step_validation(func):
    """ The :func:`on_step_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_step_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback with :meth:`~.Callback.on_step_validation` calling func
    """
    return bind_to(Callback.on_step_validation)(func)


def add_to_loss(func):
    """ The :func:`add_to_loss` decorator is used to initialise a :class:`.Callback` with the value returned from func
    being added to the loss

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        :class:`.Callback`: Initialised callback which adds the returned value from func to the loss
    """
    @on_criterion
    @on_criterion_validation
    def add_to_loss_callback(state):
        state[torchbearer.LOSS] = state[torchbearer.LOSS] + func(state)

    return add_to_loss_callback


def once(fcn):
    """
    Decorator to fire a callback once in the entire fitting procedure.

    Args:
        fcn (function): the `torchbearer callback` function to decorate.

    Returns:
        the decorator
    """
    d = {'done': False}

    def _once(_):
        if not d['done']:
            d['done'] = True
            return True
        return False

    return only_if(_once)(fcn)


def once_per_epoch(fcn):
    """
    Decorator to fire a callback once (on the first call) in any given epoch.

    Args:
        fcn (function): the `torchbearer callback` function to decorate.

    Returns:
        the decorator
    """
    d = {'last_epoch': None}

    def ope(state):
        if state[torchbearer.EPOCH] != d['last_epoch']:
            d['last_epoch'] = state[torchbearer.EPOCH]
            return True
        return False

    return only_if(ope)(fcn)


def only_if(condition_expr):
    """
    Decorator to fire a callback only if the given conditional expression function returns True.

    Args:
        condition_expr: a function/lambda that must evaluate to true for the\
                        decorated `torchbearer callback` to be called. The `state`\
                        object passed to the callback will be passed as an argument\
                        to the condition function.

    Returns:
        the decorator
    """
    def condition_decorator(fcn):
        if isinstance(fcn, LambdaCallback):
            fcn.func = condition_decorator(fcn.func)
            return fcn
        else:
            count = count_args(fcn)
            if count == 2:  # Assume Class method
                def decfcn(o, state):
                    if condition_expr(state):
                        fcn(o, state)
            else:  # Assume function of state
                def decfcn(state):
                    if condition_expr(state):
                        fcn(state)
            return decfcn
    return condition_decorator



import torchbearer

from torchbearer.callbacks import Callback

import torch


class GradientNormClipping(Callback):
    """GradientNormClipping callback, which uses 'torch.nn.utils.clip_grad_norm\_' to clip the gradient norms to the
    given value. If params is None they will be retrieved from state.

    Args:
        max_norm (float or int): max norm of the gradients
        norm_type (float or int): type of the used p-norm. Can be ``'inf'`` for
            infinity norm.
        params (Iterable[Tensor] or Tensor, optional): an iterable of Tensors or a
            single Tensor that will have gradients normalized, otherwise this is retrieved from state

    State Requirements:
        - :attr:`torchbearer.state.MODEL`: Model should have the `parameters` method
    """
    def __init__(self, max_norm, norm_type=2, params=None):
        super(GradientNormClipping, self).__init__()

        self.max_norm = max_norm
        self.norm_type = norm_type
        self.params = params

    def on_start(self, state):
        """If params is None then retrieve from the model.

        Args:
            state (dict): The :class:`.Trial` state
        """
        if self.params is None:
            self.params = filter(lambda p: p.requires_grad, state[torchbearer.MODEL].parameters())

    def on_backward(self, state):
        """Between the backward pass (which computes the gradients) and the step call (which updates the parameters),
        clip the gradient.

        Args:
            state (dict): The :class:`.Trial` state
        """
        torch.nn.utils.clip_grad_norm_(self.params, self.max_norm, norm_type=self.norm_type)


class GradientClipping(Callback):
    """GradientClipping callback, which uses 'torch.nn.utils.clip_grad_value\_' to clip the gradients of the given
    parameters to the given value. If params is None they will be retrieved from state.

    Args:
        clip_value (float or int): maximum allowed value of the gradients
            The gradients are clipped in the range [-clip_value, clip_value]
        params (Iterable[Tensor] or Tensor, optional): an iterable of Tensors or a
            single Tensor that will have gradients normalized, otherwise this is retrieved from state

    State Requirements:
        - :attr:`torchbearer.state.MODEL`: Model should have the `parameters` method
    """
    def __init__(self, clip_value, params=None):

        super(GradientClipping, self).__init__()

        self.clip_value = clip_value
        self.params = params

    def on_start(self, state):
        """If params is None then retrieve from the model.

        Args:
            state (dict): The :class:`.Trial` state
        """
        if self.params is None:
            self.params = filter(lambda p: p.requires_grad, state[torchbearer.MODEL].parameters())

    def on_backward(self, state):
        """Between the backward pass (which computes the gradients) and the step call (which updates the parameters),
        clip the gradient.

        Args:
            state (dict): The :class:`.Trial` state
        """
        torch.nn.utils.clip_grad_value_(self.params, self.clip_value)



import os
from unittest import TestCase

import torch
import torch.nn as nn
from mock import patch, Mock, ANY

import torchbearer
from torchbearer.callbacks import TensorBoard, TensorBoardImages, TensorBoardProjector, TensorBoardText


class TestTensorBoard(TestCase):
    @patch('tensorboardX.SummaryWriter')
    @patch('visdom.Visdom')
    @patch('torchbearer.callbacks.tensor_board.os.path.isdir')
    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    def test_get_writer_oserror(self, mockdirs, isdir, _, __):
        from torchbearer.callbacks.tensor_board import get_writer
        import sys

        isdir.return_value = True
        mockdirs.side_effect = OSError

        self.assertRaises(OSError, lambda: get_writer('test', 'nothing', visdom=True))
        if sys.version_info[0] >= 3:
            mockdirs.assert_called_once_with('test', exist_ok=True)
        else:
            mockdirs.assert_called_once_with('test')

    @patch('tensorboardX.SummaryWriter')
    @patch('visdom.Visdom')
    @patch('torchbearer.callbacks.tensor_board.os.path.isdir')
    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    def test_get_writer_oserror_eexist(self, mockdirs, isdir, _, __):
        from torchbearer.callbacks.tensor_board import get_writer
        import sys
        import errno

        class MyError(OSError):
            def __init__(self):
                self.errno = errno.EEXIST

        isdir.return_value = True
        mockdirs.side_effect = MyError

        get_writer('test', 'nothing', visdom=True)
        if sys.version_info[0] >= 3:
            mockdirs.assert_called_once_with('test', exist_ok=True)
        else:
            mockdirs.assert_called_once_with('test')

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_log_dir(self, mock_board, _):
        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoard(write_epoch_metrics=False)
        tboard.on_start(state)
        tboard.on_end(state)

        mock_board.assert_called_once_with(log_dir=os.path.join('./logs', 'Sequential_torchbearer'))

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_log_dir_visdom(self, mock_visdom, mock_writer, _):
        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}
        mock_writer.__delete__ = Mock()

        tboard = TensorBoard(visdom=True, write_epoch_metrics=False)

        tboard.on_start(state)
        tboard.on_end(state)

        self.assertEqual(mock_visdom.call_count, 1)
        self.assertTrue(mock_visdom.call_args[1]['log_to_filename'] == os.path.join('./logs', 'Sequential_torchbearer',
                                                                                    'log.log'))

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_batch_log_dir(self, mock_board, _):
        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.EPOCH: 0}

        tboard = TensorBoard(write_batch_metrics=True, write_graph=False, write_epoch_metrics=False)
        tboard.on_start(state)
        tboard.on_start_epoch(state)
        tboard.on_end_epoch(state)
        tboard.on_end(state)

        mock_board.assert_called_with(log_dir=os.path.join('./logs', 'Sequential_torchbearer', 'epoch-0'))

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_batch_log_dir_visdom(self, mock_visdom, mock_writer, _):
        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)),
                 torchbearer.EPOCH: 0, torchbearer.METRICS: {'test': 1}, torchbearer.BATCH: 0}

        tboard = TensorBoard(visdom=True, write_batch_metrics=True, write_graph=False, write_epoch_metrics=False)
        tboard.on_start(state)
        tboard.on_start_epoch(state)
        tboard.on_end_epoch(state)
        tboard.on_end(state)

        self.assertTrue(mock_visdom.call_args[1]['log_to_filename'] == os.path.join('./logs', 'Sequential_torchbearer', 'epoch', 'log.log'))

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    @patch('torch.rand')
    def test_write_graph(self, mock_rand, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_graph = Mock()
        mock_rand.return_value = 1

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.X: torch.zeros(1, 1, 9, 9)}

        tboard = TensorBoard(write_epoch_metrics=False)
        tboard.on_start(state)
        tboard.on_sample(state)
        tboard.on_end(state)

        mock_rand.assert_called_once_with(state[torchbearer.X].size(), requires_grad=False)
        self.assertEqual(mock_board.return_value.add_graph.call_count, 1)
        self.assertEqual(str(state[torchbearer.MODEL]), str(mock_board.return_value.add_graph.call_args_list[0][0][0]))
        self.assertNotEqual(state[torchbearer.MODEL], mock_board.return_value.add_graph.call_args_list[0][0][0])

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_writer_closed_on_end(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.close = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoard(write_epoch_metrics=False)
        tboard.on_start(state)
        tboard.on_end({})
        self.assertEqual(mock_board.return_value.close.call_count, 1)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_writer_closed_on_end_visdom(self, mock_visdom, mock_writer, _):
        mock_writer.return_value = Mock()
        mock_writer.return_value.close = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoard(visdom=True, write_epoch_metrics=False)
        tboard.on_start(state)
        tboard.on_end({})
        self.assertEqual(mock_writer.return_value.close.call_count, 1)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_batch_writer_closed_on_end_epoch(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.close = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.EPOCH: 0}

        tboard = TensorBoard(write_batch_metrics=True, write_epoch_metrics=False)
        tboard.on_start(state)
        tboard.on_start_epoch(state)
        tboard.on_end_epoch({})
        self.assertEqual(mock_board.return_value.close.call_count, 1)
        tboard.on_end(state)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_batch_writer_closed_on_end_epoch_visdom(self, mock_visdom, mock_writer, _):
        mock_writer.return_value = Mock()
        mock_writer.return_value.close = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.EPOCH: 0}

        tboard = TensorBoard(visdom=True, write_batch_metrics=True, write_epoch_metrics=False)
        tboard.on_start(state)
        tboard.on_start_epoch(state)
        tboard.on_end_epoch({})
        tboard.on_end(state)
        self.assertTrue(mock_writer.return_value.close.call_count == 2)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_batch_metrics(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_scalar = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)),
                 torchbearer.EPOCH: 0, torchbearer.METRICS: {'test': 1}, torchbearer.BATCH: 0}

        tboard = TensorBoard(write_batch_metrics=True, write_epoch_metrics=False)
        tboard.on_start(state)
        tboard.on_start_epoch(state)
        tboard.on_step_training(state)
        mock_board.return_value.add_scalar.assert_called_once_with('batch/test', 1, 0)
        mock_board.return_value.add_scalar.reset_mock()
        tboard.on_step_validation(state)
        mock_board.return_value.add_scalar.assert_called_once_with('batch/test', 1, 0)
        tboard.on_end_epoch(state)
        tboard.on_end(state)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_batch_metrics_visdom(self, mock_visdom, mock_writer, _):
        mock_writer.return_value = Mock()
        mock_writer.return_value.add_scalar = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)),
                 torchbearer.EPOCH: 0, torchbearer.METRICS: {'test': 1}, torchbearer.BATCH: 0, torchbearer.TRAIN_STEPS: 0}

        tboard = TensorBoard(visdom=True, write_batch_metrics=True, write_epoch_metrics=False)
        tboard.on_start(state)
        tboard.on_start_epoch(state)
        tboard.on_step_training(state)
        mock_writer.return_value.add_scalar.assert_called_once_with('test', 1, 0, main_tag='batch')
        mock_writer.return_value.add_scalar.reset_mock()
        tboard.on_step_validation(state)
        mock_writer.return_value.add_scalar.assert_called_once_with('test', 1, 0, main_tag='batch')
        tboard.on_end_epoch(state)
        tboard.on_end(state)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_epoch_metrics(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_scalar = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.EPOCH: 0,
                 torchbearer.METRICS: {'test': 1}}

        tboard = TensorBoard(write_batch_metrics=False, write_epoch_metrics=True)
        tboard.on_start(state)
        tboard.on_start_epoch(state)
        tboard.on_end_epoch(state)
        mock_board.return_value.add_scalar.assert_called_once_with('epoch/test', 1, 0)
        tboard.on_end(state)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_epoch_metrics_visdom(self, mock_visdom, mock_writer, _):
        mock_writer.return_value = Mock()
        mock_writer.return_value.add_scalar = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.EPOCH: 0,
                 torchbearer.METRICS: {'test': 1}}

        tboard = TensorBoard(visdom=True, write_batch_metrics=False, write_epoch_metrics=True)
        tboard.on_start(state)
        tboard.on_start_epoch(state)
        tboard.on_end_epoch(state)
        mock_writer.return_value.add_scalar.assert_called_once_with('test', 1, 0, main_tag='epoch')
        tboard.on_end(state)


class TestTensorBoardImages(TestCase):
    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_log_dir(self, mock_board, _):
        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages(log_dir='./test', comment='torchbearer')
        tboard.on_start(state)
        tboard.on_end(state)

        mock_board.assert_called_once_with(log_dir=os.path.join('./test', 'Sequential_torchbearer'))

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_log_dir_visdom(self, mock_visdom, mock_writer, _):
        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}
        mock_writer.__delete__ = Mock()

        tboard = TensorBoardImages(visdom=True, log_dir='./test', comment='torchbearer')

        tboard.on_start(state)
        tboard.on_end(state)

        self.assertEqual(mock_visdom.call_count, 1)
        self.assertTrue(mock_visdom.call_args[1]['log_to_filename'] == os.path.join('./test', 'Sequential_torchbearer',
                                                                                    'log.log'))

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_writer_closed_on_end(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.close = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages()
        tboard.on_start(state)
        tboard.on_end({})
        self.assertEqual(mock_board.return_value.close.call_count, 1)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_writer_closed_on_end_visdom_visdom(self, mock_visdom, mock_writer, _):
        mock_writer.return_value = Mock()
        mock_writer.return_value.close = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoard(visdom=True)
        tboard.on_start(state)
        tboard.on_end({})
        self.assertEqual(mock_writer.return_value.close.call_count, 1)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('torchbearer.callbacks.tensor_board.utils.make_grid')
    @patch('tensorboardX.SummaryWriter')
    def test_simple_case(self, mock_board, mock_grid, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_image = Mock()

        mock_grid.return_value = 10

        state = {'x': torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 1,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages(name='test', key='x', write_each_epoch=False, num_images=18, nrow=9, padding=3,
                                   normalize=True, norm_range='tmp', scale_each=True, pad_value=1)

        tboard.on_start(state)
        tboard.on_step_validation(state)

        mock_grid.assert_called_once_with(ANY, nrow=9, padding=3, normalize=True, range='tmp', scale_each=True,
                                          pad_value=1)
        mock_board.return_value.add_image.assert_called_once_with('test', 10, 1)
        self.assertTrue(mock_grid.call_args[0][0].size() == state['x'].size())
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.utils.make_grid')
    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_simple_case_visdom(self, mock_visdom, mock_writer, _, mock_grid):
        mock_writer.return_value = Mock()
        mock_writer.return_value.add_image = Mock()

        mock_grid.return_value = 10

        state = {'x': torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 1,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages(visdom=True, name='test', key='x', write_each_epoch=False, num_images=18, nrow=9, padding=3,
                                   normalize=True, norm_range='tmp', scale_each=True, pad_value=1)

        tboard.on_start(state)
        tboard.on_step_validation(state)

        mock_grid.assert_called_once_with(ANY, nrow=9, padding=3, normalize=True, range='tmp', scale_each=True,
                                          pad_value=1)
        mock_writer.return_value.add_image.assert_called_once_with('test1', 10, 1)
        self.assertTrue(mock_grid.call_args[0][0].size() == state['x'].size())
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('torchbearer.callbacks.tensor_board.utils.make_grid')
    @patch('tensorboardX.SummaryWriter')
    def test_multi_batch(self, mock_board, mock_grid, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_image = Mock()

        mock_grid.return_value = 10

        state = {'x': torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 1,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages(name='test', key='x', write_each_epoch=False, num_images=36, nrow=9, padding=3,
                                   normalize=True, norm_range='tmp', scale_each=True, pad_value=1)

        tboard.on_start(state)
        tboard.on_step_validation(state)
        tboard.on_step_validation(state)

        mock_grid.assert_called_once_with(ANY, nrow=9, padding=3, normalize=True, range='tmp', scale_each=True,
                                          pad_value=1)
        mock_board.return_value.add_image.assert_called_once_with('test', 10, 1)
        self.assertTrue(mock_grid.call_args[0][0].size() == torch.ones(36, 3, 10, 10).size())
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.utils.make_grid')
    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_multi_batch_visdom(self, mock_visdom, mock_writer, _, mock_grid):
        mock_writer.return_value = Mock()
        mock_writer.return_value.add_image = Mock()

        mock_grid.return_value = 10

        state = {'x': torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 1,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages(visdom=True, name='test', key='x', write_each_epoch=False, num_images=36, nrow=9, padding=3,
                                   normalize=True, norm_range='tmp', scale_each=True, pad_value=1)

        tboard.on_start(state)
        tboard.on_step_validation(state)
        tboard.on_step_validation(state)

        mock_grid.assert_called_once_with(ANY, nrow=9, padding=3, normalize=True, range='tmp', scale_each=True,
                                          pad_value=1)
        mock_writer.return_value.add_image.assert_called_once_with('test1', 10, 1)
        self.assertTrue(mock_grid.call_args[0][0].size() == torch.ones(36, 3, 10, 10).size())
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('torchbearer.callbacks.tensor_board.utils.make_grid')
    @patch('tensorboardX.SummaryWriter')
    def test_multi_epoch(self, mock_board, mock_grid, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_image = Mock()

        mock_grid.return_value = 10

        state = {'x': torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 1,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages(name='test', key='x', write_each_epoch=True, num_images=36, nrow=9, padding=3,
                                   normalize=True, norm_range='tmp', scale_each=True, pad_value=1)

        tboard.on_start(state)
        tboard.on_step_validation(state)
        tboard.on_end_epoch(state)
        tboard.on_step_validation(state)

        mock_grid.assert_called_once_with(ANY, nrow=9, padding=3, normalize=True, range='tmp', scale_each=True,
                                          pad_value=1)
        mock_board.return_value.add_image.assert_called_once_with('test', 10, 1)
        self.assertTrue(mock_grid.call_args[0][0].size() == torch.ones(36, 3, 10, 10).size())
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.utils.make_grid')
    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_multi_epoch_visdom(self, mock_visdom, mock_writer, _, mock_grid):
        mock_writer.return_value = Mock()
        mock_writer.return_value.add_image = Mock()

        mock_grid.return_value = 10

        state = {'x': torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 1,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages(visdom=True, name='test', key='x', write_each_epoch=True, num_images=36, nrow=9, padding=3,
                                   normalize=True, norm_range='tmp', scale_each=True, pad_value=1)

        tboard.on_start(state)
        tboard.on_step_validation(state)
        tboard.on_end_epoch(state)
        tboard.on_step_validation(state)

        mock_grid.assert_called_once_with(ANY, nrow=9, padding=3, normalize=True, range='tmp', scale_each=True,
                                          pad_value=1)
        mock_writer.return_value.add_image.assert_called_once_with('test1', 10, 1)
        self.assertTrue(mock_grid.call_args[0][0].size() == torch.ones(36, 3, 10, 10).size())
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('torchbearer.callbacks.tensor_board.utils.make_grid')
    @patch('tensorboardX.SummaryWriter')
    def test_single_channel(self, mock_board, mock_grid, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_image = Mock()

        mock_grid.return_value = 10

        state = {'x': torch.ones(18, 10, 10), torchbearer.EPOCH: 1,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages(name='test', key='x', write_each_epoch=True, num_images=18, nrow=9, padding=3,
                                   normalize=True, norm_range='tmp', scale_each=True, pad_value=1)

        tboard.on_start(state)
        tboard.on_step_validation(state)

        mock_grid.assert_called_once_with(ANY, nrow=9, padding=3, normalize=True, range='tmp', scale_each=True,
                                          pad_value=1)
        mock_board.return_value.add_image.assert_called_once_with('test', 10, 1)
        self.assertTrue(mock_grid.call_args[0][0].size() == torch.ones(18, 1, 10, 10).size())
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.utils.make_grid')
    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_single_channel_visdom(self, mock_visdom, mock_writer, _, mock_grid):
        mock_writer.return_value = Mock()
        mock_writer.return_value.add_image = Mock()

        mock_grid.return_value = 10

        state = {'x': torch.ones(18, 10, 10), torchbearer.EPOCH: 1,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages(visdom=True, name='test', key='x', write_each_epoch=True, num_images=18, nrow=9, padding=3,
                                   normalize=True, norm_range='tmp', scale_each=True, pad_value=1)

        tboard.on_start(state)
        tboard.on_step_validation(state)

        mock_grid.assert_called_once_with(ANY, nrow=9, padding=3, normalize=True, range='tmp', scale_each=True,
                                          pad_value=1)
        mock_writer.return_value.add_image.assert_called_once_with('test1', 10, 1)
        self.assertTrue(mock_grid.call_args[0][0].size() == torch.ones(18, 1, 10, 10).size())
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('torchbearer.callbacks.tensor_board.utils.make_grid')
    @patch('tensorboardX.SummaryWriter')
    def test_odd_batches(self, mock_board, mock_grid, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_image = Mock()

        mock_grid.return_value = 10

        state = {'x': torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 1,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages(name='test', key='x', write_each_epoch=True, num_images=40, nrow=9, padding=3,
                                   normalize=True, norm_range='tmp', scale_each=True, pad_value=1)

        tboard.on_start(state)
        tboard.on_step_validation(state)
        tboard.on_step_validation(state)
        tboard.on_step_validation(state)

        mock_grid.assert_called_once_with(ANY, nrow=9, padding=3, normalize=True, range='tmp', scale_each=True,
                                          pad_value=1)
        mock_board.return_value.add_image.assert_called_once_with('test', 10, 1)
        self.assertTrue(mock_grid.call_args[0][0].size() == torch.ones(40, 3, 10, 10).size())
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.utils.make_grid')
    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_odd_batches_visdom(self, mock_visdom, mock_writer, _, mock_grid):
        mock_writer.return_value = Mock()
        mock_writer.return_value.add_image = Mock()

        mock_grid.return_value = 10

        state = {'x': torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 1,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardImages(visdom=True, name='test', key='x', write_each_epoch=True, num_images=40, nrow=9, padding=3,
                                   normalize=True, norm_range='tmp', scale_each=True, pad_value=1)

        tboard.on_start(state)
        tboard.on_step_validation(state)
        tboard.on_step_validation(state)
        tboard.on_step_validation(state)

        mock_grid.assert_called_once_with(ANY, nrow=9, padding=3, normalize=True, range='tmp', scale_each=True,
                                          pad_value=1)
        mock_writer.return_value.add_image.assert_called_once_with('test1', 10, 1)
        self.assertTrue(mock_grid.call_args[0][0].size() == torch.ones(40, 3, 10, 10).size())
        tboard.on_end({})


class TestTensorBoardProjector(TestCase):
    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_log_dir(self, mock_board, _):
        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardProjector(log_dir='./test', comment='torchbearer')
        tboard.on_start(state)
        tboard.on_end(state)

        mock_board.assert_called_once_with(log_dir=os.path.join('./test', 'Sequential_torchbearer'))

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_writer_closed_on_end(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.close = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3))}

        tboard = TensorBoardProjector()
        tboard.on_start(state)
        tboard.on_end({})
        self.assertEqual(mock_board.return_value.close.call_count, 1)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_simple_case(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_embedding = Mock()

        state = {torchbearer.X: torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 0,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.Y_TRUE: torch.ones(18),
                 torchbearer.BATCH: 0}

        tboard = TensorBoardProjector(num_images=18, avg_data_channels=False, write_data=False,
                                      features_key=torchbearer.Y_TRUE)

        tboard.on_start(state)
        tboard.on_step_validation(state)

        mock_board.return_value.add_embedding.assert_called_once_with(ANY, metadata=ANY, label_img=ANY, tag='features',
                                                                      global_step=0)
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[0][0].size() == state[torchbearer.Y_TRUE].unsqueeze(
                1).size())
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['metadata'].size() == state[torchbearer.Y_TRUE].size())
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['label_img'].size() == state[torchbearer.X].size())
        tboard.on_end(state)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_multi_epoch(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_embedding = Mock()

        state = {torchbearer.X: torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 0,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.Y_TRUE: torch.ones(18),
                 torchbearer.BATCH: 0}

        tboard = TensorBoardProjector(num_images=18, avg_data_channels=False, write_data=False,
                                      features_key=torchbearer.Y_TRUE)

        tboard.on_start(state)
        tboard.on_step_validation(state)

        mock_board.return_value.add_embedding.assert_called_once_with(ANY, metadata=ANY, label_img=ANY, tag='features',
                                                                      global_step=0)
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[0][0].size() == state[torchbearer.Y_TRUE].unsqueeze(
                1).size())
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['metadata'].size() == state[torchbearer.Y_TRUE].size())
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['label_img'].size() == state[torchbearer.X].size())

        tboard.on_end_epoch({})
        mock_board.return_value.add_embedding.reset_mock()

        tboard.on_step_validation(state)

        mock_board.return_value.add_embedding.assert_called_once_with(ANY, metadata=ANY, label_img=ANY, tag='features',
                                                                      global_step=0)
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[0][0].size() == state[torchbearer.Y_TRUE].unsqueeze(
                1).size())
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['metadata'].size() == state[torchbearer.Y_TRUE].size())
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['label_img'].size() == state[torchbearer.X].size())
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_multi_batch(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_embedding = Mock()

        state = {torchbearer.X: torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 0,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.Y_TRUE: torch.ones(18),
                 torchbearer.BATCH: 0}

        tboard = TensorBoardProjector(num_images=45, avg_data_channels=False, write_data=False,
                                      features_key=torchbearer.Y_TRUE)

        tboard.on_start(state)
        for i in range(3):
            state[torchbearer.BATCH] = i
            tboard.on_step_validation(state)

        mock_board.return_value.add_embedding.assert_called_once_with(ANY, metadata=ANY, label_img=ANY, tag='features',
                                                                      global_step=0)
        self.assertTrue(mock_board.return_value.add_embedding.call_args[0][0].size() == torch.Size([45, 1]))
        self.assertTrue(mock_board.return_value.add_embedding.call_args[1]['metadata'].size() == torch.Size([45]))
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['label_img'].size() == torch.Size([45, 3, 10, 10]))
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_multi_batch_data(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_embedding = Mock()

        state = {torchbearer.X: torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 0,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.Y_TRUE: torch.ones(18),
                 torchbearer.BATCH: 0}

        tboard = TensorBoardProjector(num_images=45, avg_data_channels=False, write_data=True, write_features=False)

        tboard.on_start(state)
        for i in range(3):
            state[torchbearer.BATCH] = i
            tboard.on_step_validation(state)

        mock_board.return_value.add_embedding.assert_called_once_with(ANY, metadata=ANY, label_img=ANY, tag='data',
                                                                      global_step=-1)
        self.assertTrue(mock_board.return_value.add_embedding.call_args[0][0].size() == torch.Size([45, 300]))
        self.assertTrue(mock_board.return_value.add_embedding.call_args[1]['metadata'].size() == torch.Size([45]))
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['label_img'].size() == torch.Size([45, 3, 10, 10]))
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_channel_average(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_embedding = Mock()

        state = {torchbearer.X: torch.ones(18, 3, 10, 10), torchbearer.EPOCH: 0,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.Y_TRUE: torch.ones(18),
                 torchbearer.BATCH: 0}

        tboard = TensorBoardProjector(num_images=18, avg_data_channels=True, write_data=True, write_features=False)

        tboard.on_start(state)
        tboard.on_step_validation(state)

        mock_board.return_value.add_embedding.assert_called_once_with(ANY, metadata=ANY, label_img=ANY, tag='data',
                                                                      global_step=-1)
        self.assertTrue(mock_board.return_value.add_embedding.call_args[0][0].size() == torch.Size([18, 100]))
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['metadata'].size() == state[torchbearer.Y_TRUE].size())
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['label_img'].size() == state[torchbearer.X].size())
        tboard.on_end({})

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_no_channels(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_embedding = Mock()

        state = {torchbearer.X: torch.ones(18, 10, 10), torchbearer.EPOCH: 0,
                 torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)), torchbearer.Y_TRUE: torch.ones(18),
                 torchbearer.BATCH: 0}

        tboard = TensorBoardProjector(num_images=18, avg_data_channels=False, write_data=True, write_features=False)

        tboard.on_start(state)
        tboard.on_step_validation(state)

        mock_board.return_value.add_embedding.assert_called_once_with(ANY, metadata=ANY, label_img=ANY, tag='data',
                                                                      global_step=-1)
        self.assertTrue(mock_board.return_value.add_embedding.call_args[0][0].size() == torch.Size([18, 100]))
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['metadata'].size() == state[torchbearer.Y_TRUE].size())
        self.assertTrue(
            mock_board.return_value.add_embedding.call_args[1]['label_img'].size() == torch.Size([18, 1, 10, 10]))
        tboard.on_end({})


class TestTensorbardText(TestCase):
    def test_table_formatter_one_metric(self):
        tf = TensorBoardText.table_formatter

        metrics = str({'test_metric_1': 1})
        table = tf(metrics).replace(" ", "")

        correct_table = '<table><th>Metric</th><th>Value</th><tr><td>test_metric_1</td><td>1</td></tr></table>'
        self.assertEqual(table, correct_table)

    def test_table_formatter_two_metrics(self):
        tf = TensorBoardText.table_formatter

        metrics = str({'test_metric_1': 1, 'test_metric_2': 2})
        table = tf(metrics).replace(" ", "")

        self.assertIn('<tr><td>test_metric_1</td><td>1</td></tr>', table)
        self.assertIn('<tr><td>test_metric_2</td><td>2</td></tr>', table)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_epoch_writer(self, mock_writer, _):
        tboard = TensorBoardText(log_trial_summary=False)

        metrics = {'test_metric_1': 1, 'test_metric_2': 1}
        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)),
                 torchbearer.EPOCH: 1, torchbearer.METRICS: metrics}
        metric_string = TensorBoardText.table_formatter(str(metrics))

        tboard.on_start(state)
        tboard.on_start_training(state)
        tboard.on_start_epoch(state)
        tboard.on_end_epoch(state)
        mock_writer.return_value.add_text.assert_called_once_with('epoch', metric_string, 1)
        tboard.on_end(state)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_epoch_writer_visdom(self, mock_visdom, mock_writer, _):
        tboard = TensorBoardText(visdom=True, log_trial_summary=False)

        metrics = {'test_metric_1': 1, 'test_metric_2': 1}
        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)),
                 torchbearer.EPOCH: 1, torchbearer.METRICS: metrics}
        metric_string = TensorBoardText.table_formatter(str(metrics))

        tboard.on_start(state)
        tboard.on_start_training(state)
        tboard.on_start_epoch(state)
        tboard.on_end_epoch(state)
        mock_writer.return_value.add_text.assert_called_once_with('epoch', '<h4>Epoch 1</h4>'+metric_string, 1)
        tboard.on_end(state)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_batch_writer(self, mock_writer, _):
        tboard = TensorBoardText(write_epoch_metrics=False, write_batch_metrics=True, log_trial_summary=False)

        metrics = {'test_metric_1': 1, 'test_metric_2': 1}
        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)),
                 torchbearer.EPOCH: 1, torchbearer.BATCH: 100, torchbearer.METRICS: metrics}
        metric_string = TensorBoardText.table_formatter(str(metrics))

        tboard.on_start(state)
        tboard.on_start_training(state)
        tboard.on_start_epoch(state)
        tboard.on_step_training(state)
        mock_writer.return_value.add_text.assert_called_once_with('batch', metric_string, 100)
        tboard.on_end_epoch(state)
        tboard.on_end(state)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_batch_writer_visdom(self, mock_visdom, mock_writer, _):
        tboard = TensorBoardText(visdom=True, write_epoch_metrics=False, write_batch_metrics=True, log_trial_summary=False)

        metrics = {'test_metric_1': 1, 'test_metric_2': 1}
        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)),
                 torchbearer.EPOCH: 1, torchbearer.BATCH: 100, torchbearer.METRICS: metrics}
        metric_string = TensorBoardText.table_formatter(str(metrics))
        metric_string = '<h3>Epoch {} - Batch {}</h3>'.format(state[torchbearer.EPOCH], state[torchbearer.BATCH])+metric_string

        tboard.on_start(state)
        tboard.on_start_training(state)
        tboard.on_start_epoch(state)
        tboard.on_step_training(state)
        mock_writer.return_value.add_text.assert_called_once_with('batch', metric_string, 1)
        tboard.on_end_epoch(state)
        tboard.on_end(state)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_batch_metrics(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_text = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)),
                 torchbearer.EPOCH: 0, torchbearer.METRICS: {'test': 1}, torchbearer.BATCH: 0}

        tboard = TensorBoardText(write_batch_metrics=True, write_epoch_metrics=False, log_trial_summary=False)
        tboard.on_start(state)
        tboard.on_start_epoch(state)
        tboard.on_step_training(state)
        mock_board.return_value.add_text.assert_called_once_with('batch', TensorBoardText.table_formatter(str(state[torchbearer.METRICS])), 0)
        mock_board.return_value.add_text.reset_mock()
        tboard.on_end_epoch(state)
        tboard.on_end(state)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.torchvis.VisdomWriter')
    @patch('visdom.Visdom')
    def test_batch_metrics_visdom(self, mock_visdom, mock_writer, _):
        mock_writer.return_value = Mock()
        mock_writer.return_value.add_text = Mock()

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)),
                 torchbearer.EPOCH: 0, torchbearer.METRICS: {'test': 1}, torchbearer.BATCH: 0, torchbearer.TRAIN_STEPS: 0}

        tboard = TensorBoardText(visdom=True, write_batch_metrics=True, write_epoch_metrics=False, log_trial_summary=False)
        tboard.on_start(state)
        tboard.on_start_epoch(state)
        tboard.on_step_training(state)
        mock_writer.return_value.add_text.assert_called_once_with('batch', '<h3>Epoch {} - Batch {}</h3>'.format(state[torchbearer.EPOCH], state[torchbearer.BATCH])+TensorBoardText.table_formatter(str(state[torchbearer.METRICS])), 1)
        mock_writer.return_value.add_text.reset_mock()
        tboard.on_step_validation(state)
        tboard.on_end(state)

    @patch('torchbearer.callbacks.tensor_board.os.makedirs')
    @patch('tensorboardX.SummaryWriter')
    def test_log_summary(self, mock_board, _):
        mock_board.return_value = Mock()
        mock_board.return_value.add_text = Mock()
        mock_self = 'test'

        state = {torchbearer.MODEL: nn.Sequential(nn.Conv2d(3, 3, 3)),
                 torchbearer.EPOCH: 0, torchbearer.METRICS: {'test': 1}, torchbearer.BATCH: 0, torchbearer.SELF: mock_self}
        tboard = TensorBoardText(write_batch_metrics=False, write_epoch_metrics=False, log_trial_summary=True)
        tboard.on_start(state)
        self.assertEqual(mock_board.return_value.add_text.call_args[0][0], 'trial')
        self.assertEqual(mock_board.return_value.add_text.call_args[0][1], str(mock_self))



from __future__ import print_function
from collections import OrderedDict
from functools import partial

from tqdm import tqdm

import torchbearer
from torchbearer.callbacks import Callback


def _format_metrics(metrics, rounder):
    # Adapted from https://github.com/tqdm/tqdm
    postfix = OrderedDict([])
    for key in sorted(metrics.keys()):
        postfix[key] = metrics[key]

    for key in postfix.keys():
        try:
            postfix[key] = str(rounder(postfix[key]))
        except TypeError:
            try:
                postfix[key] = str(list(map(rounder, postfix[key])))
            except TypeError:
                postfix[key] = str(postfix[key])
    postfix = ', '.join(key + '=' + postfix[key].strip() for key in postfix.keys())
    return postfix


class ConsolePrinter(Callback):
    """The ConsolePrinter callback simply outputs the training metrics to the console.

    Args:
        validation_label_letter (str): This is the letter displayed after the epoch number indicating the current phase
            of training
        precision (int): Precision of the number format in decimal places

    State Requirements:
        - :attr:`torchbearer.state.EPOCH`: The current epoch number
        - :attr:`torchbearer.state.MAX_EPOCHS`: The total number of epochs for this run
        - :attr:`torchbearer.state.BATCH`: The current batch / iteration number
        - :attr:`torchbearer.state.STEPS`: The total number of steps / batches / iterations for this epoch
        - :attr:`torchbearer.state.METRICS`: The metrics dict to print
    """
    def __init__(self, validation_label_letter='v', precision=4):
        super(ConsolePrinter, self).__init__()
        self.validation_label = validation_label_letter
        self.rounder = partial(round, ndigits=precision)

    def _step(self, state, letter, steps):
        epoch_str = '{:d}/{:d}({:s}): '.format(state[torchbearer.EPOCH], state[torchbearer.MAX_EPOCHS], letter)
        batch_str = '{:d}/{:d} '.format(state[torchbearer.BATCH], steps)
        stats_str = _format_metrics(state[torchbearer.METRICS], self.rounder)
        print('\r' + epoch_str + batch_str + stats_str, end='')

    def _end(self, state, letter):
        epoch_str = '{:d}/{:d}({:s}): '.format(state[torchbearer.EPOCH], state[torchbearer.MAX_EPOCHS], letter)
        stats_str = _format_metrics(state[torchbearer.METRICS], self.rounder)
        print('\r' + epoch_str + stats_str)

    def on_step_training(self, state):
        self._step(state, 't', state[torchbearer.STEPS])

    def on_end_training(self, state):
        self._end(state, 't')

    def on_step_validation(self, state):
        self._step(state, self.validation_label, state[torchbearer.STEPS])

    def on_end_validation(self, state):
        self._end(state, self.validation_label)


class Tqdm(Callback):
    """The Tqdm callback outputs the progress and metrics for training and validation loops to the console using TQDM.
    The given key is used to label validation output.

    Args:
        tqdm_module: The tqdm module to use. If none, defaults to tqdm or tqdm_notebook if in notebook
        validation_label_letter (str): The letter to use for validation outputs.
        precision (int): Precision of the number format in decimal places
        on_epoch (bool): If True, output a single progress bar which tracks epochs
        tqdm_args: Any extra keyword args provided here will be passed through to the tqdm module constructor.
            See `github.com/tqdm/tqdm#parameters <https://github.com/tqdm/tqdm#parameters>`_ for more details.

    State Requirements:
        - :attr:`torchbearer.state.EPOCH`: The current epoch number
        - :attr:`torchbearer.state.MAX_EPOCHS`: The total number of epochs for this run
        - :attr:`torchbearer.state.STEPS`: The total number of steps / batches / iterations for this epoch
        - :attr:`torchbearer.state.METRICS`: The metrics dict to print
        - :attr:`torchbearer.state.HISTORY`: The history of the :class:`.Trial` object
    """
    def __init__(self, tqdm_module=None, validation_label_letter='v', precision=4, on_epoch=False, **tqdm_args):
        if torchbearer.magics.is_notebook() and tqdm_module is None:
            from tqdm import tqdm_notebook
            self.tqdm_module = tqdm_notebook
        else:
            self.tqdm_module = tqdm if tqdm_module is None else tqdm_module

        self._loader = None
        self.validation_label = validation_label_letter
        self.rounder = partial(round, ndigits=precision)
        self._on_epoch = on_epoch
        self.tqdm_args = tqdm_args

    def _on_start(self, state, letter):
        bar_desc = '{:d}/{:d}({:s})'.format(state[torchbearer.EPOCH], state[torchbearer.MAX_EPOCHS], letter)
        self._loader = self.tqdm_module(total=state[torchbearer.STEPS], desc=bar_desc, **self.tqdm_args)

    def _update(self, state):
        self._loader.update(1)
        self._loader.set_postfix_str(_format_metrics(state[torchbearer.METRICS], self.rounder))

    def _close(self, state):
        self._loader.set_postfix_str(_format_metrics(state[torchbearer.METRICS], self.rounder))
        self._loader.close()

    def on_start(self, state):
        if self._on_epoch:

            n = len(state[torchbearer.HISTORY])
            self._loader = self.tqdm_module(initial=n, total=state[torchbearer.MAX_EPOCHS], **self.tqdm_args)

            if n > 0:
                metrics = state[torchbearer.HISTORY][-1][1]
                state[torchbearer.METRICS] = metrics
                self._update(state)

    def on_end_epoch(self, state):
        if self._on_epoch:
            self._update(state)

    def on_end(self, state):
        if self._on_epoch:
            self._close(state)

    def on_start_training(self, state):
        """Initialise the TQDM bar for this training phase.

        Args:
            state (dict): The :class:`.Trial` state
        """
        if not self._on_epoch:
            self._on_start(state, 't')

    def on_step_training(self, state):
        """Update the bar with the metrics from this step.

        Args:
            state (dict): The :class:`.Trial` state
        """
        if not self._on_epoch:
            self._update(state)

    def on_end_training(self, state):
        """Update the bar with the terminal training metrics and then close.

        Args:
            state (dict): The :class:`.Trial` state
        """
        if not self._on_epoch:
            self._close(state)

    def on_start_validation(self, state):
        """Initialise the TQDM bar for this validation phase.

        Args:
            state (dict): The :class:`.Trial` state
        """
        if not self._on_epoch:
            self._on_start(state, self.validation_label)

    def on_step_validation(self, state):
        """Update the bar with the metrics from this step.

        Args:
            state (dict): The :class:`.Trial` state
        """
        if not self._on_epoch:
            self._update(state)

    def on_end_validation(self, state):
        """Update the bar with the terminal validation metrics and then close.

        Args:
            state (dict): The :class:`.Trial` state
        """
        if not self._on_epoch:
            self._close(state)



import unittest

import torch
from torch.nn import Module

import torchbearer as tb


class Net(Module):
    def __init__(self, x):
        super(Net, self).__init__()
        self.pars = torch.nn.Parameter(x)

    def f(self):
        """
        function to be minimised:
        f(x) = (x[0]-5)^2 + x[1]^2 + (x[2]-1)^2
        Solution:
        x = [5,0,1]
        """
        out = torch.zeros_like(self.pars)
        out[0] = self.pars[0]-5
        out[1] = self.pars[1]
        out[2] = self.pars[2]-1
        return torch.sum(out**2)

    def forward(self, _):
        return self.f()


class NetWithState(Net):
    def forward(self, _, state=None):
        if state is None:
            raise ValueError
        return super(NetWithState, self).forward(_)


def loss(y_pred, y_true):
    return y_pred


class TestEndToEnd(unittest.TestCase):
    def test_basic_opt(self):
        p = torch.tensor([2.0, 1.0, 10.0])
        training_steps = 1000

        model = NetWithState(p)
        optim = torch.optim.SGD(model.parameters(), lr=0.01)

        tbmodel = tb.Trial(model, optim, loss).for_train_steps(training_steps).for_val_steps(1)
        tbmodel.run()

        self.assertAlmostEqual(model.pars[0].item(), 5.0, places=4)
        self.assertAlmostEqual(model.pars[1].item(), 0.0, places=4)
        self.assertAlmostEqual(model.pars[2].item(), 1.0, places=4)

    def test_basic_checkpoint(self):
        p = torch.tensor([2.0, 1.0, 10.0])
        training_steps = 500

        model = Net(p)
        optim = torch.optim.SGD(model.parameters(), lr=0.01)

        tbmodel = tb.Trial(model, optim, loss, callbacks=[tb.callbacks.MostRecent(filepath='test.pt')]).for_train_steps(training_steps).for_val_steps(1)
        tbmodel.run(2)  # Simulate 2 'epochs'

        # Reload
        p = torch.tensor([2.0, 1.0, 10.0])
        model = Net(p)
        optim = torch.optim.SGD(model.parameters(), lr=0.01)

        tbmodel = tb.Trial(model, optim, loss, callbacks=[tb.callbacks.MostRecent(filepath='test.pt')]).for_train_steps(training_steps)
        tbmodel.load_state_dict(torch.load('test.pt'))
        self.assertEqual(len(tbmodel.state[tb.HISTORY]), 2)
        self.assertAlmostEqual(model.pars[0].item(), 5.0, places=4)
        self.assertAlmostEqual(model.pars[1].item(), 0.0, places=4)
        self.assertAlmostEqual(model.pars[2].item(), 1.0, places=4)

        import os
        os.remove('test.pt')

    def test_only_model(self):
        p = torch.tensor([2.0, 1.0, 10.0])

        model = Net(p)

        tbmodel = tb.Trial(model)
        self.assertListEqual(tbmodel.run(), [])

    def test_no_model(self):
        tbmodel = tb.Trial(None)

        import warnings
        with warnings.catch_warnings(record=True) as w:
            tbmodel.run()
            self.assertTrue(len(w) == 1)



import functools

import torch
import torchbearer
from torchbearer import cite
import torchbearer.callbacks as callbacks

beta_vae = """
@article{higgins2016beta,
  title={beta-vae: Learning basic visual concepts with a constrained variational framework},
  author={Higgins, Irina and Matthey, Loic and Pal, Arka and Burgess, Christopher and Glorot, Xavier and Botvinick, Matthew and Mohamed, Shakir and Lerchner, Alexander},
  year={2016}
}
"""

understanding_beta_vae = """
@article{burgess2018understanding,
  title={Understanding disentangling in beta-vae},
  author={Burgess, Christopher P and Higgins, Irina and Pal, Arka and Matthey, Loic and Watters, Nick and Desjardins, Guillaume and Lerchner, Alexander},
  journal={arXiv preprint arXiv:1804.03599},
  year={2018}
}
"""

weibullKL="""
@article{DBLP:journals/corr/Bauckhage14,
  author    = {Christian Bauckhage},
  title     = {Computing the Kullback-Leibler Divergence between two Generalized
               Gamma Distributions},
  journal   = {CoRR},
  volume    = {abs/1401.6853},
  year      = {2014}
}
"""


class DivergenceBase(callbacks.Callback):
    """The :class:`DivergenceBase` class is an abstract base class which defines a series of useful methods for dealing
    with divergences. The keys dict given on init is used to map objects in state to kwargs in the compute function.

    Args:
        keys (dict): Dictionary which maps kwarg names to :class:`.StateKey` objects. When :meth:`compute` is called,
            the given kwargs are mapped to their associated values in state.
        state_key: If not None, the value outputted by :meth:`compute` is stored in state with the given key.
    """

    def __init__(self, keys, state_key=None):
        self.keys = keys
        self.state_key = state_key

        self._post = lambda loss: loss
        self._reduce = lambda x: x.sum(1).mean(0)

        def store(state, val):
            state[state_key] = val.detach()

        self._store = store if state_key is not None else (lambda state, val: None)

    def with_post_function(self, post_fcn):
        """Register the given post function, to be applied after to loss after reduction.

        Args:
            post_fcn: A function of loss which applies some operation (e.g. multiplying by beta)

        Returns:
            :class:`.Divergence`: self
        """
        old_post = self._post
        self._post = lambda loss: post_fcn(old_post(loss))
        return self

    def compute(self, **kwargs):
        """Compute the loss with the given kwargs defined in the constructor.

        Args:
            kwargs: The bound kwargs, taken from state with the keys given in the constructor

        Returns:
            The calculated divergence as a two dimensional tensor (batch, distribution dimensions)
        """
        raise NotImplementedError

    def loss(self, state):
        kwargs = dict([(name, state[self.keys[name]]) for name in self.keys.keys()])
        return self.compute(**kwargs)

    def on_criterion(self, state):
        div = self._reduce(self.loss(state))
        self._store(state, div)
        state[torchbearer.LOSS] = state[torchbearer.LOSS] + self._post(div)

    def on_criterion_validation(self, state):
        div = self._reduce(self.loss(state))
        self._store(state, div)
        state[torchbearer.LOSS] = state[torchbearer.LOSS] + self._post(div)

    def with_reduction(self, reduction_fcn):
        """Override the reduction operation with the given function, use this if your divergence doesn't output a two
        dimensional tensor.

        Args:
            reduction_fcn: The function to be applied to the divergence output and return a single value

        Returns:
            :class:`.Divergence`: self
        """
        self._reduce = reduction_fcn
        return self

    def with_sum_mean_reduction(self):
        """Override the reduction function to take a sum over dimension one and a mean over dimension zero. (default)

        Returns:
            :class:`.Divergence`: self
        """
        return self.with_reduction(lambda x: x.sum(1).mean(0))

    def with_sum_sum_reduction(self):
        """Override the reduction function to take a sum over all dimensions.

        Returns:
            :class:`.Divergence`: self
        """
        return self.with_reduction(lambda x: x.sum())

    @cite(beta_vae)
    def with_beta(self, beta):
        """Multiply the divergence by the given beta, as introduced by beta-vae.

        Args:
            beta (float): The beta (> 1) to multiply by.

        Returns:
            :class:`.Divergence`: self
        """
        def beta_div(loss):
            return beta * loss
        return self.with_post_function(beta_div)

    @cite(understanding_beta_vae)
    def with_linear_capacity(self, min_c=0, max_c=25, steps=100000, gamma=1000):
        """Limit divergence by capacity, linearly increased from min_c to max_c for steps, as introduced in
        `Understanding disentangling in beta-VAE`.

        Args:
            min_c (float): Minimum capacity
            max_c (float): Maximum capacity
            steps (int): Number of steps to increase over
            gamma (float): Multiplicative gamma, usually a high number

        Returns:
            :class:`.Divergence`: self
        """
        inc = steps / (max_c - min_c)
        d = {'c': min_c}
        old_callback = self.on_step_training

        @functools.wraps(old_callback)
        def step_c(state):
            if d['c'] < max_c:
                d['c'] += inc
            return old_callback(state)
        self.on_step_training = step_c

        def limit_div(loss):
            return gamma * (loss - d['c']).abs()
        return self.with_post_function(limit_div)


class SimpleNormalUnitNormalKL(DivergenceBase):
    """A KL divergence between a SimpleNormal (or similar) instance and a fixed unit normal (N[0, 1]) target.

    .. note::

       The distribution object must have mu and logvar attributes

    Args:
        input_key: :class:`.StateKey` instance which will be mapped to the distribution object.
        state_key: If not None, the value outputted by :meth:`compute` is stored in state with the given key.
    """
    def __init__(self, input_key, state_key=None):
        super(SimpleNormalUnitNormalKL, self).__init__({'input': input_key}, state_key=state_key)

    def compute(self, input):
        mu, logvar = input.mu, input.logvar
        return 0.5 * (mu.pow(2) + logvar.exp() - logvar - 1)


class SimpleNormalSimpleNormalKL(DivergenceBase):
    """A KL divergence between two SimpleNormal (or similar) distributions.

    .. note::

       The distribution objects must have mu and logvar attributes

    Args:
        input_key: :class:`.StateKey` instance which will be mapped to the input distribution object.
        target_key: :class:`.StateKey` instance which will be mapped to the target distribution object.
        state_key: If not None, the value outputted by :meth:`compute` is stored in state with the given key.
    """
    def __init__(self, input_key, target_key, state_key=None):
        super(SimpleNormalSimpleNormalKL, self).__init__({'input': input_key, 'target': target_key}, state_key=state_key)

    def compute(self, input, target):
        mu_1, logvar_1 = input.mu, input.logvar
        mu_2, logvar_2 = target.mu, target.logvar
        return 0.5 * (logvar_1.exp() / logvar_2.exp() + (mu_2 - mu_1).pow(2) / logvar_2.exp() + logvar_2 - logvar_1 - 1)

@cite(weibullKL)
class SimpleWeibullSimpleWeibullKL(DivergenceBase):
    """A KL divergence between two SimpleWeibull (or similar) distributions.

    .. note::

       The distribution object must have lambda (scale) and k (shape) attributes

     Args:
        input_key: :class:`.StateKey` instance which will be mapped to the input distribution object.
        target_key: :class:`.StateKey` instance which will be mapped to the target distribution object.
        state_key: If not None, the value outputted by :meth:`compute` is stored in state with the given key.
    """
    def __init__(self, input_key, target_key, state_key=None):
        super(SimpleWeibullSimpleWeibullKL, self).__init__({'input': input_key, 'target': target_key}, state_key=state_key)
        self.gamma=0.5772

    def compute(self, input, target):
        lambda_1, k_1 = input.l, input.k
        lambda_2, k_2 = target.l, target.k
        a = torch.log(k_1 / torch.pow(lambda_1, k_1))
        b = torch.log(k_2 / torch.pow(lambda_2, k_2))
        c = torch.mul((k_1 - k_2), (torch.log(lambda_1) - self.gamma / k_1))
        n = k_2 / k_1 + 1
        gammaf = torch.exp(torch.lgamma(n))
        d = torch.mul(torch.pow(torch.div(lambda_1, lambda_2), k_2), gammaf)
        loss = torch.mean(a - b + c + d - 1)
        return loss

class SimpleExponentialSimpleExponentialKL(DivergenceBase):
    """A KL divergence between two SimpleExponential (or similar) distributions.

    .. note::

       The distribution object must have lograte attribute

     Args:
        input_key: :class:`.StateKey` instance which will be mapped to the input distribution object.
        target_key: :class:`.StateKey` instance which will be mapped to the target distribution object.
        state_key: If not None, the value outputted by :meth:`compute` is stored in state with the given key.
    """
    def __init__(self, input_key, target_key, state_key=None):
        super(SimpleExponentialSimpleExponentialKL, self).__init__({'input': input_key, 'target': target_key}, state_key=state_key)

    def compute(self, input, target):
        lograte_1 = input.lograte
        lograte_2 = target.lograte
        loss = lograte_1 - lograte_2 + lograte_2.exp()/lograte_1.exp() - 1
        return loss




import unittest

from mock import Mock

import torchbearer
from torchbearer.cv_utils import *


class TestCVUtils(unittest.TestCase):

    def test_train_valid_splitter_sizes(self):
        x = range(1, 101)
        y = range(1, 101)

        x = torch.Tensor(x)
        y = torch.Tensor(y)

        valid_split = 0.1
        shuffle = False

        x, y, x_val, y_val = train_valid_splitter(x, y, valid_split, shuffle)
        self.assertTrue(x.size()[0] == 90)
        self.assertTrue(y.size()[0] == 90)
        self.assertTrue(x_val.size()[0] == 10)
        self.assertTrue(y_val.size()[0] == 10)

    def test_train_valid_splitter_sizes_2(self):
        x = range(1, 101)
        y = range(1, 101)

        x = torch.Tensor(x)
        y = torch.Tensor(y)

        valid_split = 0.4
        shuffle = False

        x, y, x_val, y_val = train_valid_splitter(x, y, valid_split, shuffle)
        self.assertTrue(x.size()[0] == 60)
        self.assertTrue(y.size()[0] == 60)
        self.assertTrue(x_val.size()[0] == 40)
        self.assertTrue(y_val.size()[0] == 40)

    def test_train_valid_splitter_sizes_2d(self):
        x = range(1, 101)
        y = range(1, 101)

        x = torch.Tensor(x)
        y = torch.Tensor(y)

        x = torch.stack([x, x], -1)
        y = torch.stack([y, y], -1)

        valid_split = 0.1
        shuffle = False

        x, y, x_val, y_val = train_valid_splitter(x, y, valid_split, shuffle)
        self.assertTrue(list(x.size()) == [90, 2])
        self.assertTrue(list(y.size()) == [90, 2])
        self.assertTrue(list(x_val.size()) == [10, 2])
        self.assertTrue(list(y_val.size()) == [10, 2])

    def test_train_valid_splitter_order(self):
        x = range(1, 101)
        y = range(1, 101)

        x = torch.Tensor(x)
        y = torch.Tensor(y)

        valid_split = 0.1
        shuffle = False

        x, y, x_val, y_val = train_valid_splitter(x, y, valid_split, shuffle)
        self.assertTrue(list(x.numpy()) == list(range(11, 101)))
        self.assertTrue(list(y.numpy()) == list(range(11, 101)))
        self.assertTrue(list(x_val.numpy()) == list(range(1, 11)))
        self.assertTrue(list(y_val.numpy()) == list(range(1, 11)))

    def test_train_valid_splitter_split_negative(self):
        x = range(1, 101)
        y = range(1, 101)

        x = torch.Tensor(x)
        y = torch.Tensor(y)

        valid_split = -0.1
        shuffle = False

        x, y, x_val, y_val = train_valid_splitter(x, y, valid_split, shuffle)
        self.assertTrue(list(x.numpy()) == list(range(91, 101)))
        self.assertTrue(list(y.numpy()) == list(range(91, 101)))
        self.assertTrue(list(x_val.numpy()) == list(range(1, 91)))
        self.assertTrue(list(y_val.numpy()) == list(range(1, 91)))

    def test_train_valid_splitter_split_zero(self):
        x = range(1, 101)
        y = range(1, 101)

        x = torch.Tensor(x)
        y = torch.Tensor(y)

        valid_split = 0
        shuffle = False

        x, y, x_val, y_val = train_valid_splitter(x, y, valid_split, shuffle)
        self.assertTrue(list(x.numpy()) == list(range(1, 101)))
        self.assertTrue(list(y.numpy()) == list(range(1, 101)))
        self.assertTrue(list(x_val.numpy()) == list(range(0, 0)))
        self.assertTrue(list(y_val.numpy()) == list(range(0, 0)))

    def test_train_valid_splitter_split_too_big(self):
        x = range(1, 101)
        y = range(1, 101)

        x = torch.Tensor(x)
        y = torch.Tensor(y)

        valid_split = 1.8
        shuffle = False

        x, y, x_val, y_val = train_valid_splitter(x, y, valid_split, shuffle)
        self.assertTrue(list(x.numpy()) == list(range(0, 0)))
        self.assertTrue(list(y.numpy()) == list(range(0, 0)))
        self.assertTrue(list(x_val.numpy()) == list(range(1, 101)))
        self.assertTrue(list(y_val.numpy()) == list(range(1, 101)))

    def test_train_valid_splitter_shuffle_size(self):
        x = range(1, 101)
        y = range(1, 101)

        x = torch.Tensor(x)
        y = torch.Tensor(y)

        valid_split = 0.1
        shuffle = True

        x, y, x_val, y_val = train_valid_splitter(x, y, valid_split, shuffle)
        self.assertTrue(x.size()[0] == 90)
        self.assertTrue(y.size()[0] == 90)
        self.assertTrue(x_val.size()[0] == 10)
        self.assertTrue(y_val.size()[0] == 10)

    def test_get_train_valid_sets_splitter_args(self):
        x = range(1, 101)
        y = range(1, 101)

        x = torch.Tensor(x)
        y = torch.Tensor(y)

        valid_split = 0.1
        shuffle = True

        torchbearer.cv_utils.train_valid_splitter = Mock(return_value=(x,y,x,y))
        tvs = torchbearer.cv_utils.train_valid_splitter

        trainset, valset = get_train_valid_sets(x, y, None, valid_split, shuffle)
        self.assertEqual(tvs.call_count, 1)
        self.assertTrue(tvs.call_args[0][-1] == valid_split)
        self.assertTrue(list(tvs.call_args[0][0].numpy()) == list(x.numpy()))
        self.assertTrue(list(tvs.call_args[0][1].numpy()) == list(y.numpy()))
        self.assertTrue(tvs.call_args[1]['shuffle'] == shuffle)

    def test_get_train_valid_sets_given_valid_data(self):
        x = range(1, 101)
        y = range(1, 101)
        x_val = range(101, 121)
        y_val = range(101, 121)

        x = torch.Tensor(x)
        y = torch.Tensor(y)
        x_val = torch.Tensor(x_val)
        y_val = torch.Tensor(y_val)

        valid_split = 0.1
        shuffle = False

        trainset, valset = get_train_valid_sets(x, y, (x_val, y_val), valid_split, shuffle)
        self.assertTrue(len(valset) == len(x_val))

    def test_get_train_valid_sets_no_valid(self):
        x = range(1, 101)
        y = range(1, 101)

        x = torch.Tensor(x)
        y = torch.Tensor(y)

        valid_split = None
        shuffle = False

        trainset, valset = get_train_valid_sets(x, y, None, valid_split, shuffle)
        self.assertTrue(valset is None)
        self.assertTrue(len(trainset) == len(x))

    def test_DatasetValidationSplitter(self):
        data = torch.Tensor(list(range(1000)))
        dataset = TensorDataset(data)

        splitter = DatasetValidationSplitter(len(dataset), 0.1)
        trainset = splitter.get_train_dataset(dataset)
        validset = splitter.get_val_dataset(dataset)

        self.assertTrue(len(trainset) == 900)
        self.assertTrue(len(validset) == 100)

        # Check for ids in both train and validation set
        collision = False
        for id in trainset:
            if id in validset.ids:
                collision = True
        self.assertFalse(collision)

    def test_DatasetValidationSplitter_seed(self):
        data = torch.Tensor(list(range(1000)))
        dataset = TensorDataset(data)

        splitter_1 = DatasetValidationSplitter(len(dataset), 0.1, shuffle_seed=1)
        trainset_1 = splitter_1.get_train_dataset(dataset)
        validset_1 = splitter_1.get_val_dataset(dataset)

        splitter_2 = DatasetValidationSplitter(len(dataset), 0.1, shuffle_seed=1)
        trainset_2 = splitter_2.get_train_dataset(dataset)
        validset_2 = splitter_2.get_val_dataset(dataset)

        self.assertTrue(trainset_1.ids[0] == trainset_2.ids[0])
        self.assertTrue(validset_1.ids[0] == validset_2.ids[0])





import unittest

import torch

import torchbearer
from torchbearer.bases import Metric, Callback
from torchbearer.metrics.primitives import CategoricalAccuracy


class TestMetric(unittest.TestCase):
    def setUp(self):
        self._state = {
            torchbearer.Y_TRUE: torch.LongTensor([0, 1, 2, 2, 1]),
            torchbearer.Y_PRED: torch.FloatTensor([
                [0.9, 0.1, 0.1], # Correct
                [0.1, 0.9, 0.1], # Correct
                [0.1, 0.1, 0.9], # Correct
                [0.9, 0.1, 0.1], # Incorrect
                [0.9, 0.1, 0.1], # Incorrect
            ])
        }
        self._state[torchbearer.Y_PRED].requires_grad = True
        self._targets = [1, 1, 1, 0, 0]
        self._metric = CategoricalAccuracy().root

    def test_requires_grad(self):
        result = self._metric.process(self._state)
        self.assertTrue(self._state[torchbearer.Y_PRED].requires_grad is True)
        self.assertTrue(result.requires_grad is False)

    def test_empty_methods(self):
        metric = Metric(name='test')
        self.assertTrue(metric.process() is None)
        self.assertTrue(metric.process_final() is None)


class TestCallback(unittest.TestCase):
    def test_state_dict(self):
        callback = Callback()

        self.assertEqual(callback.state_dict(), {})
        self.assertEqual(callback.load_state_dict({}), callback)

    def test_str(self):
        callback = Callback()
        self.assertEqual(str(callback).strip(), "torchbearer.bases.Callback")

    def test_empty_methods(self):
        callback = Callback()

        self.assertIsNone(callback.on_start({}))
        self.assertIsNone(callback.on_start_epoch({}))
        self.assertIsNone(callback.on_start_training({}))
        self.assertIsNone(callback.on_sample({}))
        self.assertIsNone(callback.on_forward({}))
        self.assertIsNone(callback.on_criterion({}))
        self.assertIsNone(callback.on_backward({}))
        self.assertIsNone(callback.on_step_training({}))
        self.assertIsNone(callback.on_end_training({}))
        self.assertIsNone(callback.on_end_epoch({}))
        self.assertIsNone(callback.on_checkpoint({}))
        self.assertIsNone(callback.on_end({}))
        self.assertIsNone(callback.on_start_validation({}))
        self.assertIsNone(callback.on_sample_validation({}))
        self.assertIsNone(callback.on_forward_validation({}))
        self.assertIsNone(callback.on_end_validation({}))
        self.assertIsNone(callback.on_step_validation({}))
        self.assertIsNone(callback.on_criterion_validation({}))



"""
The decorator API is the core way to interact with metrics in torchbearer. All of the classes and functionality handled
here can be reproduced by manually interacting with the classes if necessary. Broadly speaking, the decorator API is
used to construct a :class:`.MetricFactory` which will build a :class:`.MetricTree` that handles data flow between
instances of :class:`.Mean`, :class:`.RunningMean`, :class:`.Std` etc.
"""
import inspect

from torchbearer.metrics import EpochLambda, BatchLambda, ToDict, Mean, MetricTree, Std, Var, RunningMean
from .metrics import add_default


def default_for_key(key, *args, **kwargs):
    """The :func:`default_for_key` decorator will register the given metric in the global metric dict
    (`metrics.DEFAULT_METRICS`) so that it can be referenced by name in instances of :class:`.MetricList` such as in the
    list given to the :class:`.torchbearer.Model`.

    Example: ::

        @default_for_key('acc')
        class CategoricalAccuracy(metrics.BatchLambda):
            ...

    Args:
        key (str): The key to use when referencing the metric
        args: Any args to pass to the underlying metric when constructed
        kwargs: Any keyword args to pass to the underlying metric when constructed
    """
    def decorator(arg):
        add_default(key, arg, *args, **kwargs)
        return arg
    return decorator


def lambda_metric(name, on_epoch=False):
    """The :func:`lambda_metric` decorator is used to convert a lambda function `y_pred, y_true` into a :class:`.Metric`
    instance. This can be used as in the following example: ::

        @metrics.lambda_metric('my_metric')
        def my_metric(y_pred, y_true):
            ... # Calculate some metric

        model = Model(metrics=[my_metric])

    Args:
        name (str): The name of the metric (e.g. 'loss')
        on_epoch (bool): If True the metric will be an instance of :class:`.EpochLambda` instead of :class:`.BatchLambda`

    Returns:
        A decorator which replaces a function with a :class:`.Metric`
    """
    def decorator(metric_function):
        if on_epoch:
            return EpochLambda(name, metric_function)
        else:
            return BatchLambda(name, metric_function)
    return decorator


def to_dict(clazz):
    """The :func:`to_dict` decorator is used to wrap either a :class:`.Metric` class or a :class:`.Metric` instance with
    a :class:`.ToDict` instance. The result is that future output will be wrapped in a `dict[name, value]`.

    Example: ::

        >>> from torchbearer import metrics

        >>> @metrics.lambda_metric('my_metric')
        ... def my_metric(y_pred, y_true):
        ...     return y_pred + y_true
        ...
        >>> my_metric.process({'y_pred':4, 'y_true':5})
        9

        >>> @metrics.to_dict
        ... @metrics.lambda_metric('my_metric')
        ... def my_metric(y_pred, y_true):
        ...     return y_pred + y_true
        ...
        >>> my_metric.process({'y_pred':4, 'y_true':5})
        {'my_metric': 9}

    Args:
        clazz: The class to *decorate*

    Returns:
        A :class:`.ToDict` instance or a :class:`.ToDict` wrapper of the given class
    """
    if inspect.isclass(clazz):
        class Wrapper(ToDict):
            __doc__ = clazz.__doc__

            def __init__(self, *args, **kwargs):
                super(Wrapper, self).__init__(clazz(*args, **kwargs))
        Wrapper.__name__ = clazz.__name__
        return Wrapper
    else:
        return ToDict(clazz)


def _wrap_and_add_to_tree(clazz, child_func):
    if inspect.isclass(clazz):
        class Wrapper(MetricTree):
            __doc__ = clazz.__doc__

            def __init__(self, *args, **kwargs):
                inner = clazz(*args, **kwargs)
                if isinstance(inner, MetricTree):
                    super(Wrapper, self).__init__(inner.root)
                    self.children = inner.children
                else:
                    super(Wrapper, self).__init__(inner)

                self.add_child(child_func(self.root))
        Wrapper.__name__ = clazz.__name__
        return Wrapper
    else:
        inner = clazz
        if not isinstance(inner, MetricTree):
            inner = MetricTree(inner)
        inner.add_child(child_func(inner))
        return inner


def mean(clazz=None, dim=None):
    """The :func:`mean` decorator is used to add a :class:`.Mean` to the :class:`.MetricTree` which will will output a
    mean value at the end of each epoch. At build time, if the inner class is not a :class:`.MetricTree`, one will be
    created. The :class:`.Mean` will also be wrapped in a :class:`.ToDict` for simplicity.

    Example: ::

        >>> import torch
        >>> from torchbearer import metrics

        >>> @metrics.mean
        ... @metrics.lambda_metric('my_metric')
        ... def metric(y_pred, y_true):
        ...     return y_pred + y_true
        ...
        >>> metric.reset({})
        >>> metric.process({'y_pred':torch.Tensor([2]), 'y_true':torch.Tensor([2])}) # 4
        {}
        >>> metric.process({'y_pred':torch.Tensor([3]), 'y_true':torch.Tensor([3])}) # 6
        {}
        >>> metric.process({'y_pred':torch.Tensor([4]), 'y_true':torch.Tensor([4])}) # 8
        {}
        >>> metric.process_final()
        {'my_metric': 6.0}

    Args:
        clazz: The class to *decorate*
        dim (int, tuple): See :class:`.Mean`

    Returns:
        A :class:`.MetricTree` with a :class:`.Mean` appended or a wrapper class that extends :class:`.MetricTree`
    """
    if clazz is None:
        def decorator(clazz):
            return mean(clazz, dim=dim)
        return decorator

    return _wrap_and_add_to_tree(clazz, lambda metric: ToDict(Mean(metric.name, dim=dim)))


def std(clazz=None, unbiased=True, dim=None):
    """The :func:`std` decorator is used to add a :class:`.Std` to the :class:`.MetricTree` which will will output a
    sample standard deviation value at the end of each epoch. At build time, if the inner class is not a
    :class:`.MetricTree`, one will be created. The :class:`.Std` will also be wrapped in a :class:`.ToDict` (with '_std'
    appended) for simplicity.

    Example: ::

        >>> import torch
        >>> from torchbearer import metrics

        >>> @metrics.std
        ... @metrics.lambda_metric('my_metric')
        ... def metric(y_pred, y_true):
        ...     return y_pred + y_true
        ...
        >>> metric.reset({})
        >>> metric.process({'y_pred':torch.Tensor([2]), 'y_true':torch.Tensor([2])}) # 4
        {}
        >>> metric.process({'y_pred':torch.Tensor([3]), 'y_true':torch.Tensor([3])}) # 6
        {}
        >>> metric.process({'y_pred':torch.Tensor([4]), 'y_true':torch.Tensor([4])}) # 8
        {}
        >>> '%.4f' % metric.process_final()['my_metric_std']
        '2.0000'

    Args:
        clazz: The class to *decorate*
        unbiased (bool): See :class:`.Std`
        dim (int, tuple): See :class:`.Std`

    Returns:
        A :class:`.MetricTree` with a :class:`.Std` appended or a wrapper class that extends :class:`.MetricTree`
    """
    if clazz is None:
        def decorator(clazz):
            return std(clazz, unbiased=unbiased, dim=dim)
        return decorator

    return _wrap_and_add_to_tree(clazz, lambda metric: ToDict(Std(metric.name + '_std', unbiased=unbiased, dim=dim)))


def var(clazz=None, unbiased=True, dim=None):
    """The :func:`var` decorator is used to add a :class:`.Var` to the :class:`.MetricTree` which will will output a
    sample variance value at the end of each epoch. At build time, if the inner class is not a :class:`.MetricTree`, one
    will be created. The :class:`.Var` will also be wrapped in a :class:`.ToDict` (with '_var' appended) for simplicity.

    Example: ::

        >>> import torch
        >>> from torchbearer import metrics

        >>> @metrics.var
        ... @metrics.lambda_metric('my_metric')
        ... def metric(y_pred, y_true):
        ...     return y_pred + y_true
        ...
        >>> metric.reset({})
        >>> metric.process({'y_pred':torch.Tensor([2]), 'y_true':torch.Tensor([2])}) # 4
        {}
        >>> metric.process({'y_pred':torch.Tensor([3]), 'y_true':torch.Tensor([3])}) # 6
        {}
        >>> metric.process({'y_pred':torch.Tensor([4]), 'y_true':torch.Tensor([4])}) # 8
        {}
        >>> '%.4f' % metric.process_final()['my_metric_var']
        '4.0000'

    Args:
        clazz: The class to *decorate*
        unbiased (bool): See :class:`.Var`
        dim (int, tuple): See :class:`.Var`

    Returns:
        A :class:`.MetricTree` with a :class:`.Var` appended or a wrapper class that extends :class:`.MetricTree`
    """
    if clazz is None:
        def decorator(clazz):
            return var(clazz, unbiased=unbiased, dim=dim)
        return decorator

    return _wrap_and_add_to_tree(clazz, lambda metric: ToDict(Var(metric.name + '_var', unbiased=unbiased, dim=dim)))


def running_mean(clazz=None, batch_size=50, step_size=10, dim=None):
    """The :func:`running_mean` decorator is used to add a :class:`.RunningMean` to the :class:`.MetricTree`. If the
    inner class is not a :class:`.MetricTree` then one will be created. The :class:`.RunningMean` will be wrapped in a
    :class:`.ToDict` (with 'running\_' prepended to the name) for simplicity.

    .. note::
        The decorator function does not need to be called if not desired, both: `@running_mean` and `@running_mean()`
        are acceptable.

    Example: ::

        >>> import torch
        >>> from torchbearer import metrics

        >>> @metrics.running_mean(step_size=2) # Update every 2 steps
        ... @metrics.lambda_metric('my_metric')
        ... def metric(y_pred, y_true):
        ...     return y_pred + y_true
        ...
        >>> metric.reset({})
        >>> metric.process({'y_pred':torch.Tensor([2]), 'y_true':torch.Tensor([2])}) # 4
        {'running_my_metric': 4.0}
        >>> metric.process({'y_pred':torch.Tensor([3]), 'y_true':torch.Tensor([3])}) # 6
        {'running_my_metric': 4.0}
        >>> metric.process({'y_pred':torch.Tensor([4]), 'y_true':torch.Tensor([4])}) # 8, triggers update
        {'running_my_metric': 6.0}

    Args:
        clazz: The class to *decorate*
        batch_size (int): See :class:`.RunningMean`
        step_size (int): See :class:`.RunningMean`
        dim (int, tuple): See :class:`.RunningMean`

    Returns:
        decorator or :class:`.MetricTree` instance or wrapper
    """
    if clazz is None:
        def decorator(clazz):
            return running_mean(clazz, batch_size=batch_size, step_size=step_size, dim=dim)
        return decorator

    return _wrap_and_add_to_tree(clazz, lambda metric: ToDict(RunningMean('running_' + metric.name, batch_size=batch_size, step_size=step_size, dim=dim)))



import unittest
from mock import Mock, patch

import torch.nn as nn
import torch.nn.functional as F

import torchbearer
from torchbearer.metrics import DefaultAccuracy


class TestDefaultAccuracy(unittest.TestCase):
    def test_defaults(self):

        state = {torchbearer.CRITERION: 'not a criterion'}
        metric = DefaultAccuracy()
        metric.reset(state)
        self.assertEqual(metric.name, 'acc')

        state = {torchbearer.CRITERION: nn.CrossEntropyLoss()}
        metric = DefaultAccuracy()
        metric.reset(state)
        self.assertEqual(metric.name, 'acc')

        state = {torchbearer.CRITERION: nn.NLLLoss()}
        metric = DefaultAccuracy()
        metric.reset(state)
        self.assertEqual(metric.name, 'acc')

        state = {torchbearer.CRITERION: F.cross_entropy}
        metric = DefaultAccuracy()
        metric.reset(state)
        self.assertEqual(metric.name, 'acc')

        state = {torchbearer.CRITERION: F.nll_loss}
        metric = DefaultAccuracy()
        metric.reset(state)
        self.assertEqual(metric.name, 'acc')

        state = {torchbearer.CRITERION: nn.MSELoss()}
        metric = DefaultAccuracy()
        metric.reset(state)
        self.assertEqual(metric.name, 'mse')

        state = {torchbearer.CRITERION: F.mse_loss}
        metric = DefaultAccuracy()
        metric.reset(state)
        self.assertEqual(metric.name, 'mse')

        state = {torchbearer.CRITERION: nn.BCELoss()}
        metric = DefaultAccuracy()
        metric.reset(state)
        self.assertEqual(metric.name, 'binary_acc')

        state = {torchbearer.CRITERION: nn.BCEWithLogitsLoss()}
        metric = DefaultAccuracy()
        metric.reset(state)
        self.assertEqual(metric.name, 'binary_acc')

        state = {torchbearer.CRITERION: F.binary_cross_entropy}
        metric = DefaultAccuracy()
        metric.reset(state)
        self.assertEqual(metric.name, 'binary_acc')

        state = {torchbearer.CRITERION: F.binary_cross_entropy_with_logits}
        metric = DefaultAccuracy()
        metric.reset(state)
        self.assertEqual(metric.name, 'binary_acc')

    @patch('torchbearer.metrics.default.CategoricalAccuracy')
    def test_pass_through(self, cat_acc):
        mock = Mock()
        cat_acc.return_value = mock
        mock.reset = Mock()
        mock.process = Mock()
        mock.process_final = Mock()
        mock.eval = Mock()
        mock.train = Mock()

        metric = DefaultAccuracy()
        metric.reset({torchbearer.CRITERION: None})
        metric.process(1, 2, 3)
        metric.process_final(4, 5, 6)
        metric.eval()
        metric.train()

        self.assertEqual(cat_acc.call_count, 1)
        mock.reset.assert_called_once_with({torchbearer.CRITERION: None})
        mock.process.assert_called_once_with(1, 2, 3)
        mock.process_final.assert_called_once_with(4, 5, 6)
        self.assertEqual(mock.eval.call_count, 1)
        self.assertEqual(mock.train.call_count, 1)

    @patch('torchbearer.metrics.default.CategoricalAccuracy')
    def test_reset_after_eval(self, cat_acc):
        metric = DefaultAccuracy()
        self.assertTrue(cat_acc.call_count == 1)
        cat_acc.reset_mock()
        metric.eval()

        state = {torchbearer.CRITERION: F.cross_entropy}
        mock = Mock()
        mock.eval = Mock()
        torchbearer.metrics.default.__loss_map__[F.cross_entropy.__name__] = lambda: mock

        metric.reset(state)

        self.assertTrue(mock.eval.call_count == 1)



"""
The state is central in torchbearer, storing all of the relevant intermediate values that may be changed or replaced
during model fitting. This module defines classes for interacting with state and all of the built in state keys used
throughout torchbearer. The :func:`state_key` function can be used to create custom state keys for use in callbacks or
metrics.

Example: ::

    from torchbearer import state_key
    MY_KEY = state_key('my_test_key')
"""
from torchbearer import Metric
import warnings

__keys__ = []


def state_key(key):
    """Computes and returns a non-conflicting key for the state dictionary when given a seed key

    Args:
        key (str): The seed key - basis for new state key

    Returns:
        :class:`.StateKey`: New state key
    """
    return StateKey(key)


class StateKey(Metric):
    """ StateKey class that is a unique state key based on the input string key. State keys are also metrics which
    retrieve themselves from state.

    Args:
        key (str): Base key
    """
    def __init__(self, key):
        self.key = self._gen_key_(key)
        super(StateKey, self).__init__(self.key)

    def process(self, state):
        return {self.name: state[self]}

    def process_final(self, state):
        return {self.name: state[self]}

    def _gen_key_(self, key):
        if key in __keys__:
            count = 1
            my_key = key + '_' + str(count)

            while my_key in __keys__:
                count += 1
                my_key = key + '_' + str(count)

            key = my_key

        __keys__.append(key)
        return key

    def __repr__(self):
        return self.key

    def __str__(self):
        return self.key

    def __eq__(self, other):
        return self.key == str(other)
        
    def __hash__(self):
        return self.key.__hash__()


class State(dict):
    """
    State dictionary that behaves like a python dict but accepts StateKeys
    """
    def __init__(self):
        super(State, self).__init__()

    def get_key(self, statekey):
        if isinstance(statekey, str):
            warnings.warn("State was accessed with a string: {}, generate keys with StateKey(str).".format(statekey), stacklevel=3)
        return statekey

    def __getitem__(self, key):
        return super(State, self).__getitem__(self.get_key(key))

    def __setitem__(self, key, val):
        super(State, self).__setitem__(self.get_key(key), val)

    def __delitem__(self, val):
        super(State, self).__delitem__(val)

    def __contains__(self, o):
        return super(State, self).__contains__(self.get_key(o))

    def update(self, d):
        new_dict = {}
        for key in d:
            new_dict[self.get_key(key)] = d[key]
        super(State, self).update(new_dict)


#: The torchbearer version
VERSION = state_key('torchbearer_version')

#: The PyTorch module / model that will be trained
MODEL = state_key('model')

#: The criterion to use when model fitting
CRITERION = state_key('criterion')

#: The optimizer to use when model fitting
OPTIMIZER = state_key('optimizer')

#: The device currently in use by the :class:`.Trial` and PyTorch model
DEVICE = state_key('device')

#: The data type of tensors in use by the model, match this to avoid type issues
DATA_TYPE = state_key('dtype')

#: The list of metrics in use by the :class:`.Trial`
METRIC_LIST = state_key('metric_list')

#: The metric dict from the current batch of data
METRICS = state_key('metrics')

#: A self refrence to the Trial object for persistence etc.
SELF = state_key('self')

#: The current epoch number
EPOCH = state_key('epoch')

#: The total number of epochs to run for
MAX_EPOCHS = state_key('max_epochs')

#: The string name of the current data
DATA = state_key('data')

#: The current data generator (DataLoader)
GENERATOR = state_key('generator')

#: The current iterator
ITERATOR = state_key('iterator')

#: The current number of steps per epoch
STEPS = state_key('steps')

#: The train data generator in the Trial object
TRAIN_GENERATOR = state_key('train_generator')

#: The number of train steps to take
TRAIN_STEPS = state_key('train_steps')

#: The flag representing train data
TRAIN_DATA = state_key('train_data')

#: Flag for refreshing of training iterator when finished instead of each epoch
INF_TRAIN_LOADING = state_key('inf_train_loading')

#: The validation data generator in the Trial object
VALIDATION_GENERATOR = state_key('validation_generator')

#: The number of validation steps to take
VALIDATION_STEPS = state_key('validation_steps')

#: The flag representing validation data
VALIDATION_DATA = state_key('validation_data')

#: The test data generator in the Trial object
TEST_GENERATOR = state_key('test_generator')

#: The number of test steps to take
TEST_STEPS = state_key('test_steps')

#: The flag representing test data
TEST_DATA = state_key('test_data')

#: A flag that can be set to true to stop the current fit call
STOP_TRAINING = state_key('stop_training')

#: The current batch of ground truth data
TARGET = Y_TRUE = state_key('y_true')

#: The current batch of predictions
PREDICTION = Y_PRED = state_key('y_pred')

#: The current batch of inputs
INPUT = X = state_key('x')

#: The sampler which loads data from the generator onto the correct device
SAMPLER = state_key('sampler')

#: The current value for the loss
LOSS = state_key('loss')

#: The key which maps to the predictions over the dataset when calling predict
FINAL_PREDICTIONS = state_key('final_predictions')

#: The current batch number
BATCH = state_key('t')

#: The timings keys used by the timer callback
TIMINGS = state_key('timings')

#: The :class:`.CallbackList` object which is called by the Trial
CALLBACK_LIST = state_key('callback_list')

#: The history list of the Trial instance
HISTORY = state_key('history')

#: The optional arguments which should be passed to the backward call
BACKWARD_ARGS = state_key('backward_args')

# Legacy
VALIDATION_ITERATOR = 'validation_iterator'
TRAIN_ITERATOR = 'train_iterator'



"""
The distributions module is an extension of the `torch.distributions` package intended to facilitate implementations
required for specific variational approaches through the :class:`.SimpleDistribution` class. Generally, using a
:class:`torch.distributions.Distribution` object should be preferred over a :class:`SimpleDistribution`, for better
argument validation and more complete implementations. However, if you need to implement something new for a specific
variational approach, then a :class:`.SimpleDistribution` may be more forgiving. Furthermore, you may find it easier
to understand the function of the implementations here.
"""
import math
from numbers import Number

import torch
from torch.distributions import Distribution
from torch.distributions.utils import broadcast_all
from torchbearer import cite

steve = """
@article{squires2019a,
title={A Variational Autoencoder for Probabilistic Non-Negative Matrix Factorisation},
author={Steven Squires and Adam Prugel-Bennett and Mahesan Niranjan},
year={2019}
}
"""


class SimpleDistribution(Distribution):
    """Abstract base class for a simple distribution which only implements rsample and log_prob. If the log_prob
    function is not differentiable with respect to the distribution parameters or the given value, then this should be
    mentioned in the documentation.
    """
    has_rsample = True

    def __init__(self, batch_shape=torch.Size(), event_shape=torch.Size()):
        super(SimpleDistribution, self).__init__(batch_shape, event_shape)

    @property
    def support(self):
        return None

    @property
    def arg_constraints(self):
        return None

    def expand(self, batch_shape, _instance=None):
        pass

    @property
    def mean(self):
        return None

    @property
    def variance(self):
        return None

    def cdf(self, value):
        pass

    def icdf(self, value):
        pass

    def enumerate_support(self, expand=True):
        pass

    def entropy(self):
        pass

    def rsample(self, sample_shape=torch.Size()):
        """
        Returns a reparameterized sample or batch of reparameterized samples if the distribution parameters are batched.
        """
        raise NotImplementedError

    def log_prob(self, value):
        """Returns the log of the probability density/mass function evaluated at `value`.
        Args:
            value (torch.Tensor, Number): Value at which to evaluate log probabilty
        """
        raise NotImplementedError


class SimpleNormal(SimpleDistribution):
    """The SimpleNormal class is a :class:`SimpleDistribution` which implements a straight forward Normal / Gaussian
    distribution. This performs significantly fewer checks than `torch.distributions.Normal`, but should be sufficient
    for the purpose of implementing a VAE.

    Args:
        mu (torch.Tensor, Number): The mean of the distribution, numbers will be cast to tensors
        logvar (torch.Tensor, Number): The log variance of the distribution, numbers will be cast to tensors
    """

    def __init__(self, mu, logvar):
        self.mu, self.logvar = broadcast_all(mu, logvar)
        if isinstance(mu, Number) and isinstance(logvar, Number):
            batch_shape = torch.Size()
        else:
            batch_shape = self.mu.size()
        super(SimpleNormal, self).__init__(batch_shape=batch_shape)

    def rsample(self, sample_shape=torch.Size()):
        """Simple rsample for a Normal distribution.

        Args:
            sample_shape (torch.Size, tuple): Shape of the sample (per mean / variance given)
        Returns:
            A reparameterized sample with gradient with respect to the distribution parameters
        """
        shape = self._extended_shape(sample_shape)
        std = self.logvar.div(2).exp_()
        eps = torch.normal(torch.zeros(shape, dtype=self.mu.dtype, device=self.mu.device),
                           torch.ones(shape, dtype=self.mu.dtype, device=self.mu.device))
        return self.mu + std * eps

    def log_prob(self, value):
        """Calculates the log probability that the given value was drawn from this distribution. Since the density of a
        Gaussian is differentiable, this function is differentiable.

        Args:
            value (torch.Tensor, Number): The sampled value
        Returns:
            The log probability that the given value was drawn from this distribution
        """
        var = self.logvar.exp()
        return - ((value - self.mu) ** 2) / (2.0 * var) - (self.logvar / 2.0) - math.log(math.sqrt(2.0 * math.pi))


class SimpleUniform(SimpleDistribution):
    """The SimpleUniform class is a :class:`SimpleDistribution` which implements a straight forward Uniform distribution
    in the interval ``[low, high)``. This performs significantly fewer checks than `torch.distributions.Uniform`, but
    should be sufficient for the purpose of implementing a VAE.

    Args:
        low (torch.Tensor, Number): The lower range of the distribution (inclusive), numbers will be cast to tensors
        high (torch.Tensor, Number): The upper range of the distribution (exclusive), numbers will be cast to tensors
    """

    def __init__(self, low, high):
        self.low, self.high = broadcast_all(low, high)
        if isinstance(low, Number) and isinstance(high, Number):
            batch_shape = torch.Size()
        else:
            batch_shape = self.low.size()
        super(SimpleUniform, self).__init__(batch_shape=batch_shape)

    def rsample(self, sample_shape=torch.Size()):
        """Simple rsample for a Uniform distribution.

        Args:
            sample_shape (torch.Size, tuple): Shape of the sample (per low / high given)
        Returns:
            A reparameterized sample with gradient with respect to the distribution parameters
        """
        shape = self._extended_shape(sample_shape)
        rand = torch.rand(shape, dtype=self.low.dtype, device=self.low.device)
        return self.low + rand * (self.high - self.low)

    def log_prob(self, value):
        """Calculates the log probability that the given value was drawn from this distribution. Since this distribution
        is uniform, the log probability is ``-log(high - low)`` for all values in the range ``[low, high)`` and -inf
        elsewhere. This function is therefore only piecewise differentiable.

        Args:
            value (torch.Tensor, Number): The sampled value
        Returns:
            The log probability that the given value was drawn from this distribution
        """
        value = value if torch.is_tensor(value) else torch.tensor(value, dtype=torch.get_default_dtype())
        lb = value.ge(self.low).type_as(self.low)
        ub = value.lt(self.high).type_as(self.low)
        return torch.log(lb.mul(ub)) - torch.log(self.high - self.low)


class SimpleExponential(SimpleDistribution):
    """The SimpleExponential class is a :class:`SimpleDistribution` which implements a straight forward Exponential
    distribution with the given lograte. This performs significantly fewer checks than `torch.distributions.Exponential`
    , but should be sufficient for the purpose of implementing a VAE. By using a lograte, the log_prob can be computed
    in a stable fashion, without taking a logarithm.

    Args:
        lograte (torch.Tensor, Number): The natural log of the rate of the distribution, numbers will be cast to tensors
    """

    def __init__(self, lograte):
        self.lograte, = broadcast_all(lograte)
        batch_shape = torch.Size() if isinstance(lograte, Number) else self.lograte.size()
        super(SimpleExponential, self).__init__(batch_shape=batch_shape)

    def rsample(self, sample_shape=torch.Size()):
        """Simple rsample for an Exponential distribution.

        Args:
            sample_shape (torch.Size, tuple): Shape of the sample (per lograte given)
        Returns:
            A reparameterized sample with gradient with respect to the distribution parameters
        """
        shape = self._extended_shape(sample_shape)
        return self.lograte.new(shape).exponential_() / self.lograte.exp()

    def log_prob(self, value):
        """Calculates the log probability that the given value was drawn from this distribution. The log_prob for this
        distribution is fully differentiable and has stable gradient since we use the lograte here.

        Args:
            value (torch.Tensor, Number): The sampled value
        Returns:
            The log probability that the given value was drawn from this distribution
        """
        return self.lograte - self.lograte.exp() * value


@cite(steve)
class SimpleWeibull(SimpleDistribution):
    """The SimpleWeibull class is a :class:`SimpleDistribution` which implements a straight forward Weibull
    distribution. This performs significantly fewer checks than `torch.distributions.Weibull`, but should be sufficient
    for the purpose of implementing a VAE.

    Args:
        l (torch.Tensor, Number): The scale parameter of the distribution, numbers will be cast to tensors
        k (torch.Tensor, Number): The shape parameter of the distribution, numbers will be cast to tensors
    """

    def __init__(self, l, k):
        self.l, self.k = broadcast_all(l, k)
        self.const=1e-8
        if isinstance(k, Number) and isinstance(l, Number):
            batch_shape = torch.Size()
        else:
            batch_shape = self.k.size()

        super(SimpleWeibull, self).__init__(batch_shape=batch_shape)

    def rsample(self, sample_shape=torch.Size()):
        """Simple rsample for a Weibull distribution.

        Args:
            sample_shape (torch.Size, tuple): Shape of the sample (per k / lambda given)
        Returns:
            A reparameterized sample with gradient with respect to the distribution parameters
        """
        shape = self._extended_shape(sample_shape)
        eps = torch.rand(shape, dtype=self.k.dtype, device=self.k.device)
        return self.l * torch.pow((-torch.log(eps)), (1/self.k))

    def log_prob(self, value):
        """Calculates the log probability that the given value was drawn from this distribution.  This function is differentiable
        and its log probability is -inf for values less than 0.

        Args:
            value (torch.Tensor, Number): The sampled value
        Returns:
            The log probability that the given value was drawn from this distribution
        """
        value = value if torch.is_tensor(value) else torch.tensor(value, dtype=torch.get_default_dtype())
        lb=value.ge(torch.zeros(value.shape, dtype=self.k.dtype, device=self.k.device)).float()
        return torch.log(lb) + torch.log(self.k/self.l) + (self.k - torch.ones(self.k.shape, dtype=self.k.dtype, device=self.k.device))*torch.log((lb*value+self.const)/self.l) - torch.pow(value/self.l, self.k)



from __future__ import print_function
import torchbearer

from torchbearer.callbacks import Callback


class EarlyStopping(Callback):
    """Callback to stop training when a monitored quantity has stopped improving.

    Args:
        monitor (str): Name of quantity in metrics to be monitored
        min_delta (float): Minimum change in the monitored quantity to qualify as an improvement, i.e. an absolute
            change of less than min_delta, will count as no improvement.
        patience (int): Number of epochs with no improvement after which training will be stopped.
        verbose (int): Verbosity mode, will print stopping info if verbose > 0
        mode (str): One of {auto, min, max}. In `min` mode, training will stop when the quantity monitored has stopped
            decreasing; in `max` mode it will stop when the quantity monitored has stopped increasing; in `auto` mode,
            the direction is automatically inferred from the name of the monitored quantity.

    State Requirements:
        - :attr:`torchbearer.state.METRICS`: Metrics should be a dict containing the given monitor key as a minimum
    """

    def __init__(self, monitor='val_loss',
                 min_delta=0, patience=0, verbose=0, mode='auto'):

        super(EarlyStopping, self).__init__()

        self.monitor = monitor
        self.min_delta = min_delta
        self.patience = patience
        self.verbose = verbose
        self.mode = mode

        if self.mode not in ['min', 'max']:
            if 'acc' in self.monitor:
                self.mode = 'max'
            else:
                self.mode = 'min'

        if self.mode == 'min':
            self.min_delta *= -1
            self.monitor_op = lambda x1, x2: x1 < x2
        elif self.mode == 'max':
            self.min_delta *= 1
            self.monitor_op = lambda x1, x2: x1 > x2

        self.wait = 0
        self.stopped_epoch = 0
        self.best = float('inf') if self.mode == 'min' else -float('inf')

    def state_dict(self):
        state_dict = {
            'wait': self.wait,
            'best': self.best,
            'stopped_epoch': self.stopped_epoch
        }
        return state_dict

    def load_state_dict(self, state_dict):
        self.wait = state_dict['wait']
        self.best = state_dict['best']
        self.stopped_epoch = state_dict['stopped_epoch']

    def on_end_epoch(self, state):
        current = state[torchbearer.METRICS][self.monitor]
        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = state[torchbearer.EPOCH]
                state[torchbearer.STOP_TRAINING] = True

    def on_end(self, state):
        if self.stopped_epoch > 0 and self.verbose > 0:
            print('Epoch %05d: early stopping' % (self.stopped_epoch + 1))



import unittest

import torchbearer
from torchbearer.state import State


class TestStateKey(unittest.TestCase):
    def test_key_metric(self):
        key = torchbearer.state_key('test')
        state = {key: 4}

        self.assertDictEqual(key.process(state), {str(key): 4})
        self.assertDictEqual(key.process_final(state), {str(key): 4})

    def test_key_repr(self):
        key = torchbearer.state_key('repr_test')
        self.assertEqual(str(key), 'repr_test')
        self.assertEqual(repr(key), 'repr_test')

    def test_key_added(self):
        key = torchbearer.state_key('key')
        self.assertTrue('key' in torchbearer.state.__keys__)

    def test_collision(self):
        _ = torchbearer.state_key('test')
        key_1 = torchbearer.state_key('test')
        key_2 = torchbearer.state_key('test')

        self.assertTrue('test' != str(key_1))
        self.assertTrue('test' != str(key_2))

    def test_duplicate_string(self):
        _ = torchbearer.state_key('test_dup')
        key_1 = torchbearer.state_key('test_dup')
        key_2 = torchbearer.state_key('test_dup')

        self.assertTrue('test_dup_1' == str(key_1))
        self.assertTrue('test_dup_2' == str(key_2))

    def test_compare_to_statekey(self):
        key_1 = torchbearer.state_key('test_compare_sk')
        key_2 = torchbearer.state_key('test_compare_sk_2')
        # Simulates same key in different sessions where the object hash is changed
        key_2.key = 'test_compare_sk'
        self.assertEqual(key_1, key_2)

    def test_compare_to_string(self):
        key_1 = torchbearer.state_key('test_compare')
        self.assertEqual(key_1, 'test_compare')


class TestState(unittest.TestCase):
    def test_contains(self):
        s = State()

        key1 = torchbearer.state_key('test_a')
        key2 = torchbearer.state_key('test_b')

        s[key1] = 1
        s[key2] = 2

        self.assertTrue(s.__contains__(key1))

    def test_delete(self):
        s = State()

        key1 = torchbearer.state_key('test_a')
        key2 = torchbearer.state_key('test_b')

        s[key1] = 1
        s[key2] = 2

        self.assertTrue(s.__contains__(key1))
        s.__delitem__(key1)
        self.assertFalse(s.__contains__(key1))

    def test_update(self):
        s = State()

        key1 = torchbearer.state_key('test_a')
        key2 = torchbearer.state_key('test_b')

        new_s = {key1: 1, key2: 2}
        s.update(new_s)

        self.assertTrue(s.__contains__(key1))
        self.assertTrue(s[key1] == 1)

    def test_update_state(self):
        s = State()
        new_s = State()

        key1 = torchbearer.state_key('test_a')
        key2 = torchbearer.state_key('test_b')

        new_s_dict = {key1: 1, key2: 2}
        new_s.update(new_s_dict)

        s.update(new_s)
        self.assertTrue(s.__contains__(key1))
        self.assertTrue(s[key1] == 1)


import torchbearer

from torchbearer.callbacks import Callback

import torch


class WeightDecay(Callback):
    """Create a WeightDecay callback which uses the given norm on the given parameters and with the given decay rate.
    If params is None (default) then the parameters will be retrieved from the model.

    Args:
        rate (float): The decay rate or lambda
        p (int): The norm level
        params (Iterable[Tensor] or Tensor, optional): an iterable of Tensors or a
            single Tensor that will have gradients normalized, otherwise this is retrieved from state

    State Requirements:
        - :attr:`torchbearer.state.MODEL`: Model should have the `parameters` method
        - :attr:`torchbearer.state.LOSS`: Loss should be a tensor that can be incremented
    """
    def __init__(self, rate=5e-4, p=2, params=None):
        super(WeightDecay, self).__init__()

        self.p = p
        self.params = params
        self.rate = rate

    def on_start(self, state):
        """Retrieve params from state['model'] if required.

        Args:
            state (dict): The :class:`.Trial` state
        """
        if self.params is None:
            self.params = state[torchbearer.MODEL].parameters()

    def on_criterion(self, state):
        """Calculate the decay term and add to state['loss'].

        Args:
            state (dict): The :class:`.Trial` state
        """
        for param in self.params:
            state[torchbearer.LOSS] += self.rate * torch.norm(param, self.p)


class L1WeightDecay(WeightDecay):
    """WeightDecay callback which uses an L1 norm with the given rate and parameters. If params is None (default) then
    the parameters will be retrieved from the model.

    Args:
        rate (float): The decay rate or lambda
        params (Iterable[Tensor] or Tensor, optional): an iterable of Tensors or a
            single Tensor that will have gradients normalized, otherwise this is retrieved from state

    State Requirements:
        - :attr:`torchbearer.state.MODEL`: Model should have the `parameters` method
        - :attr:`torchbearer.state.LOSS`: Loss should be a tensor that can be incremented
    """
    def __init__(self, rate=5e-4, params=None):
        super(L1WeightDecay, self).__init__(rate=rate, p=1, params=params)


class L2WeightDecay(WeightDecay):
    """WeightDecay callback which uses an L2 norm with the given rate and parameters. If params is None (default) then
    the parameters will be retrieved from the model.

    Args:
        rate (float): The decay rate or lambda
        params (Iterable[Tensor] or Tensor, optional): an iterable of Tensors or a
            single Tensor that will have gradients normalized, otherwise this is retrieved from state

    State Requirements:
        - :attr:`torchbearer.state.MODEL`: Model should have the `parameters` method
        - :attr:`torchbearer.state.LOSS`: Loss should be a tensor that can be incremented
    """
    def __init__(self, rate=5e-4, params=None):
        super(L2WeightDecay, self).__init__(rate=rate, p=2, params=params)
